"""
parsing.py

Provides X12 segment parsing support.
"""
import inspect
import logging
import re
from abc import ABC
from collections import defaultdict
from functools import lru_cache, wraps
from importlib import import_module
from typing import Callable, Dict, List, Optional, Set

from x12.models import X12SegmentGroup, X12Segment, X12Delimiters
from x12.segments import SEGMENT_LOOKUP

logger = logging.getLogger(__name__)

# naming convention for parsing functions:
# parse_nm1_segment, parse_st_segment, etc
PARSING_FUNCTION_REGEX = "set\\_(.)*\\_loop"


def match(segment_name: str, conditions: Dict = None) -> Callable:
    """
    The match decorator matches a X12 segment to a decorated function.

    Decorated functions are expected to have the signature: func(context: X12ParserContext).

    The following example matches a NM1 (entity name) segment:

    @match("NM1")
    def the_func(context, segment_data):
       # evaluation occurs here

    The match decorator accepts additional criteria, or conditions, for segment matching. The criteria evaluates a
    matched segment fields using equality checks.

    The following examples matches a NM1 segment where the entity_type_qualifier field is equal to "2",
    or a non-person entity.

    @match("NM1", conditions={"entity_type_qualifier": "2"})
    def the_func(context, segment_data):
       # evaluation occurs here

    Finally, the decorator adds a "segment grouping" attribute to the decorated function to optimize lookups.

    :param segment_name: The X12 segment name
    :param conditions: Dictionary of conditions. Key = x12 segment field name, Value = field value
    """

    conditions = conditions or {}

    def decorator(f):
        # add a segment grouping attribute for the function
        f.segment_group = segment_name.upper()

        @wraps(f)
        def wrapped(segment_data: Dict, data_context: X12ParserContext):
            """
            Executes the wrapped function if segment_data matches the segment_name and optional conditions.

            :param segment_data: List of segment fields.
            :param data_context: The X12ParserContext object
            """
            is_matched = segment_name.upper() == segment_data["segment_name"].upper()

            if not is_matched:
                return
            elif is_matched and not conditions:
                return f(data_context, segment_data)
            else:
                for k, v in conditions.items():
                    # evaluate "one of" matches
                    if isinstance(v, list):
                        for i in v:
                            if segment_data[k].upper() == i.upper():
                                return f(data_context, segment_data)
                    else:
                        # evaluate single matches
                        if segment_data[k].upper() == v.upper():
                            return f(data_context, segment_data)

        return wrapped

    return decorator


class X12ParserContext:
    """
    The X12ParserContext maintains the current parsing position and provides a record cache.

    The parsing position is set using `set_loop_context`, which sets the current loop name and the data structure,
    usually a dictionary, for the loop.

    Cached records include:
    * The current subscriber record
    * The current patient record
    * The current hierarchical segment

    The current hierarchical segment is set in the core parser. Subscriber and patient records are set manually within a
    transaction's loop parser.

    Cached record usage varies based on transaction type.
    """

    @property
    def loop_name(self):
        """return the current loop name"""
        return None if not self.parsed_loops else self.parsed_loops[-1]

    def set_loop_context(self, loop_name: str, loop_container: Dict) -> None:
        """
        Sets the current loop context.

        The loop context includes a list of parsed loops and the loop "container". The "loop container" is expected to
        reference the larger transactional data structure.

        :param loop_name: The "current" loop name.
        :param loop_container: The "current" loop record.
        """

        self.parsed_loops.append(loop_name)
        self.loop_container = loop_container

    def reset_transaction(self) -> None:
        """Resets transactional attributes, including loop attributes."""

        self.parsed_loops.clear()
        self.loop_container.clear()
        self.transaction_data.clear()
        self.transaction_data["header"] = {}
        self.transaction_data["footer"] = {}
        self.is_transaction_complete = False
        self.subscriber_record = None
        self.patient_record = None
        self.hl_segment = None

    def mark_transaction_complete(self) -> None:
        """Marks the current transaction as complete"""

        self.is_transaction_complete = True

    def __init__(self) -> None:
        """
        Configures the X12ParseContext instance.
        The current transaction data record is initialized to include the header and footer keys used in all transaction
        models.
        """

        self.parsed_loops: List[str] = []
        self.loop_container: Optional[Dict] = {}
        self.transaction_data: Optional[Dict] = {"header": {}, "footer": {}}
        self.is_transaction_complete: bool = False
        self.subscriber_record: Optional[Dict] = None
        self.patient_record: Optional[Dict] = None
        self.hl_segment: Optional[Dict] = None


class X12Parser(ABC):
    """
    Parses X12 segments into a transactional model.
    """

    # parsing functions used to create loops in the transaction data model
    loop_parsers = None

    def _lookup_segment_model(self, segment_name) -> X12Segment:
        """
        Searches for a segment model by name.

        :param segment_name: The segment name
        :return: The segment model
        :raises: KeyError if the segment is not found
        """

        try:
            segment_model: Optional[X12Segment] = self._segment_models[segment_name]
        except KeyError:
            msg: str = f"Unsupported segment {segment_name}"
            logger.exception(msg)
            raise
        else:
            return segment_model

    def _get_segment_field_names(self, segment_name) -> List:
        """
        Returns the field names for a specific segment

        :param segment_name: The segment name
        """

        segment_model: X12Segment = self._lookup_segment_model(segment_name)
        # exclude "meta" fields such as delimiters that are not in the segment data
        field_names = [f for f in segment_model.__fields__ if f != "delimiters"]
        return field_names

    def _get_multivalue_fields(self, segment_name) -> Dict[str, str]:
        """
        Returns a dictionary of multi-value fields.
        Dictionary key is the field name, the value is the field separator.
        This dictionary is used to split fields appropriately before they are added to the transaction model.

        :param segment_name: The segment name
        :returns: a dictionary containing the multivalue fields and field separator
        """

        multivalue_fields: Dict = {}

        segment_model: X12Segment = self._lookup_segment_model(segment_name)

        for field_name, model_data in segment_model.__fields__.items():
            if field_name == "delimiters":
                continue

            if "typing.List" in str(model_data.outer_type_):
                is_component_field: bool = model_data.field_info.extra.get(
                    "is_component", False
                )
                if is_component_field:
                    multivalue_fields[field_name] = self._delimiters.component_separator
                else:
                    multivalue_fields[
                        field_name
                    ] = self._delimiters.repetition_separator

        return multivalue_fields

    def _parse_segment(self, segment_name: str, segment_fields: List[str]) -> Dict:
        """
        Parses a X12 segment field list into a data record (dictionary) with field names and values.
        The segment_field values are assigned to field names using the pydantic model for the transaction set.

        :param segment_name: The name of the X12 segment
        :param segment_fields: List of segment field values
        :return: The parsed data as a dictionary.
        """

        segment_data: Dict = {"delimiters": self._delimiters.dict()}
        field_names: List = self._get_segment_field_names(segment_name)
        multivalue_fields: Dict = self._get_multivalue_fields(segment_name)

        # segment_fields and field_name lengths may not match
        # drive the mapping with segment_fields as it includes all fields within the transactional context
        # field_names includes ALL available fields within the specification
        for index, value in enumerate(segment_fields):
            field_name: str = field_names[index]

            multivalue_separator: str = multivalue_fields.get(field_name)
            if multivalue_separator:
                value = value.split(multivalue_separator)

            segment_data[field_name] = value if value else None

        return segment_data

    def _is_final_segment(self, segment_name) -> bool:
        """Returns True if the segment name is the final segment in a transaction set."""

        return segment_name.lower() == "se"

    def load_model(self) -> X12SegmentGroup:
        """
        Loads the instance's data record into a X12 transactional model.

        :return: The X12 transactional model.
        """

        return self._transaction_model(**self._context.transaction_data)

    def parse(
        self, segment_name: str, segment_fields: List[str]
    ) -> Optional[X12SegmentGroup]:
        """
        Parses an X12 segment into the instance's data record attribute.

        Sets the following context attributes:
        * hl_segment

        Returns a X12 Transaction Model when all segments are received.

        :param segment_name: The name of the X12 segment (ST, HL, NM1, etc)
        :param segment_fields: List of segment field values
        :return: The X12 Transaction Model if ready, otherwise None.
        """

        model: Optional[X12SegmentGroup] = None

        # convert segment data to a dictionary "record"
        segment_data: Dict = self._parse_segment(segment_name, segment_fields)

        if segment_name.lower() == "hl":
            self._context.hl_segment = segment_data

        for loop_parser in self._loop_parsers[segment_name]:
            loop_parser(segment_data, self._context)

        segment_key = f"{segment_name.lower()}_segment"

        # write segment data
        # check for existing list entries for repeating segments
        existing_value = self._context.loop_container.get(segment_key)
        if existing_value is None:
            self._context.loop_container[segment_key] = segment_data
        elif isinstance(existing_value, list):
            self._context.loop_container[segment_key].append(segment_data)

        # close transaction set and return the model
        if self._is_final_segment(segment_name):
            self._context.mark_transaction_complete()
            model: X12SegmentGroup = self.load_model()
            self._context.reset_transaction()
            return model

    def __init__(
        self,
        transaction_model: X12SegmentGroup,
        loop_parsers: Dict,
        x12_delimiters: Optional[X12Delimiters] = None,
    ) -> None:
        """
        Configures the parser instance

        :param transaction_model: The X12 transaction model.
        :param loop_parsers: The parser functions used to identify loop boundaries in the X12 transaction.
        :param x12_delimiters: The delimiters used to parse segments and fields.
        """
        self._transaction_model: X12SegmentGroup = transaction_model
        self._loop_parsers: Dict = loop_parsers
        self._delimiters: X12Delimiters = x12_delimiters or X12Delimiters()

        self._context: X12ParserContext = X12ParserContext()
        self._segment_models: Dict = SEGMENT_LOOKUP


@lru_cache
def _load_loop_parsers(transaction_code: str, implementation_version) -> Dict:
    """Returns dictionary of Loop Parsers indexed by segment name"""
    pattern = re.compile(PARSING_FUNCTION_REGEX)
    loop_parsers = defaultdict(list)

    parsing_module = import_module(
        f"x12.transactions.x12_{transaction_code}_{implementation_version}.parsing"
    )

    funcs = [
        value
        for name, value in inspect.getmembers(parsing_module, inspect.isfunction)
        if pattern.match(name)
    ]

    for f in funcs:
        loop_parsers[f.segment_group].append(f)

    return loop_parsers


@lru_cache
def _load_transaction_model(
    transaction_code: str, implementation_version: str
) -> X12SegmentGroup:
    """Returns the transaction model for the x12 transaction"""
    transaction_module = import_module(
        f"x12.transactions.x12_{transaction_code}_{implementation_version}.transaction_set"
    )

    # return transaction set model
    for _, class_value in inspect.getmembers(transaction_module, inspect.isclass):
        if hasattr(class_value, "schema"):
            props: Set = set(class_value.schema()["properties"])
            # transaction set models have header and footer attributes
            if props.issuperset({"header", "footer"}):
                return class_value


def create_parser(
    transaction_code: str, implementation_version: str, x12_delimiters: X12Delimiters
) -> X12Parser:
    """
    Creates a versioned X12 transaction parser.

    The parser includes "loop parsers" which initialize data structures for new loops as needed.

    :param transaction_code: The X12 transaction code (270, 271, 834, etc).
    :param implementation_version: The X12 implementation version (005010X279A1).
    :param x12_delimiters: The delimiters used to parse segments and fields.
    :raises: ModuleNotFoundError if the transaction module does not exist.
    :raises: ValueError if the transaction module exists, but a parser class cannot be found.
    """

    transaction_model: X12SegmentGroup = _load_transaction_model(
        transaction_code, implementation_version
    )
    loop_parsers: Dict = _load_loop_parsers(transaction_code, implementation_version)
    return X12Parser(transaction_model, loop_parsers, x12_delimiters)



"""
models.py

Base models for X12 parsing and validation.
"""
import abc
import datetime
from enum import Enum
from typing import List
from decimal import Decimal

from pydantic import BaseModel, Field


class X12Delimiters(BaseModel):
    """
    X12Delimiters models the message delimiters used within a X12 transaction.
    """

    element_separator: str = Field("*", min_length=1, max_length=1)
    repetition_separator: str = Field("^", min_length=1, max_length=1)
    segment_terminator: str = Field("~", min_length=1, max_length=1)
    component_separator: str = Field(":", min_length=1, max_length=1)

    class Config:
        # the model is immutable and hashable
        allow_mutation = False
        frozen = True


class X12SegmentName(str, Enum):
    """
    Supported X12 Segment Names
    """

    AAA = "AAA"
    AMT = "AMT"
    BHT = "BHT"
    BPR = "BPR"
    CAS = "CAS"
    CLM = "CLM"
    CLP = "CLP"
    CL1 = "CL1"
    CN1 = "CN1"
    CR1 = "CR1"
    CR2 = "CR2"
    CR3 = "CR3"
    CRC = "CRC"
    CTP = "CTP"
    CUR = "CUR"
    DMG = "DMG"
    DTM = "DTM"
    DTP = "DTP"
    EB = "EB"
    EQ = "EQ"
    FRM = "FRM"
    GE = "GE"
    GS = "GS"
    HCP = "HCP"
    HI = "HI"
    HL = "HL"
    HSD = "HSD"
    IEA = "IEA"
    III = "III"
    INS = "INS"
    ISA = "ISA"
    K3 = "K3"
    LE = "LE"
    LIN = "LIN"
    LQ = "LQ"
    LS = "LS"
    LX = "LX"
    MEA = "MEA"
    MIA = "MIA"
    MOA = "MOA"
    MPI = "MPI"
    MSG = "MSG"
    N1 = "N1"
    N3 = "N3"
    N4 = "N4"
    NM1 = "NM1"
    NTE = "NTE"
    OI = "OI"
    PAT = "PAT"
    PER = "PER"
    PLB = "PLB"
    PRV = "PRV"
    PS1 = "PS1"
    PWK = "PWK"
    QTY = "QTY"
    RDM = "RDM"
    REF = "REF"
    SBR = "SBR"
    SE = "SE"
    ST = "ST"
    STC = "STC"
    SVC = "SVC"
    SV1 = "SV1"
    SV2 = "SV2"
    SV5 = "SV5"
    SVD = "SVD"
    TRN = "TRN"
    TS2 = "TS2"
    TS3 = "TS3"


class X12Segment(abc.ABC, BaseModel):
    """
    X12BaseSegment serves as the abstract base class for all X12 segment models.
    """

    delimiters: X12Delimiters = X12Delimiters()
    segment_name: X12SegmentName

    class Config:
        """
        Default configuration for X12 Models
        """

        use_enum_values = True
        extra = "forbid"

    def _process_multivalue_field(self, field_name: str, field_value: List) -> str:
        """
        Converts a X12 multi-value (list) field into a a single delimited string.

        A "multi-value" field is a field which contains sub-fields, or components, or allows repeats.
        The X12 specification uses separate delimiters for component and repeating fields.

        :param field_name: The field name used to lookup field metadata.
        :param field_value: The field's list values
        """

        is_component_field: bool = self.__fields__[field_name].field_info.extra.get(
            "is_component", False
        )
        if is_component_field:
            join_character = self.delimiters.component_separator
        else:
            join_character = self.delimiters.repetition_separator
        return join_character.join(field_value)

    def x12(self) -> str:
        """
        :return: the X12 representation of the model instance
        """

        x12_values = []
        for k, v in self.dict(exclude={"delimiters"}).items():
            if isinstance(v, str):
                x12_values.append(v)
            elif isinstance(v, list):
                x12_values.append(self._process_multivalue_field(k, v))
            elif isinstance(v, datetime.datetime):
                x12_values.append(v.strftime("%Y%m%d%H%M"))
            elif isinstance(v, datetime.date):
                x12_values.append(v.strftime("%Y%m%d"))
            elif isinstance(v, datetime.time):
                x12_values.append(v.strftime("%H%M"))
            elif isinstance(v, Decimal):
                x12_values.append("{:.2f}".format(v))
            elif v is None:
                x12_values.append("")
            else:
                x12_values.append(str(v))

        x12_str = self.delimiters.element_separator.join(x12_values).rstrip(
            self.delimiters.element_separator
        )
        return x12_str + self.delimiters.segment_terminator


class X12SegmentGroup(abc.ABC, BaseModel):
    """
    Abstract base class for a container, typically a loop or transaction, which groups x12 segments.
    """

    def x12(self, use_new_lines=True) -> str:
        """
        :return: Generates a X12 representation of the loop using its segments.
        """
        x12_segments: List[str] = []
        fields = [f for f in self.__fields__.values() if hasattr(f.type_, "x12")]

        for f in fields:
            field_instance = getattr(self, f.name)

            if field_instance is None:
                continue
            elif isinstance(field_instance, list):
                for item in field_instance:
                    if isinstance(item, X12Segment):
                        x12_segments.append(item.x12())
                    else:
                        x12_segments.append(item.x12(use_new_lines=use_new_lines))
            else:
                if isinstance(field_instance, X12Segment):
                    x12_segments.append(field_instance.x12())
                else:
                    x12_segments.append(field_instance.x12(use_new_lines=use_new_lines))

        join_char: str = "\n" if use_new_lines else ""
        return join_char.join(x12_segments)



"""
loops.py

Models the loops, or logical segment groupings, for the HealthCare Claims Professional 837 005010X222A2 transaction set.
The HealthCare Claims Professional set organizes loops into a hierarchical and nested model.

-- Header
    -- Loop 1000A (Submitter Name)
        -- Loop 1000B (Receiver Name)
    -- Loop 2000A (Billing Provider)
        -- Loop 2010AA (Billing Provider Name)
        -- Loop 2010AB (Pay to Address)
        -- Loop 2010AC (Pay to Plan Name)
        -- Loop 2000B (Subscriber)
            -- Loop 2010BA (Subscriber Name)
            -- Loop 2010BB (Payer Name)
        -- Loop 2000C (Patient Loop)
            -- Loop 2010CA (Patient Name)
            -- Loop 2300 (Claim Information)
                -- Loop 2310A (Referring Provider Name)
                -- Loop 2310B (Rendering Provider Name)
                -- Loop 2310C (Service Facility Location Name)
                -- Loop 2310D (Supervising Provider Name)
                -- Loop 2310E (Ambulance Pickup Location)
                -- Loop 2310F (Ambulance Dropoff Location)
                -- Loop 2320 (Other Subscriber Information)
                    -- Loop 2330A (Other Subscriber Name)
                    -- Loop 2330B (Other Payer Name)
                    -- Loop 2330C (Other Payer Referring Provider Name)
                    -- Loop 2330D (Other Payer Rendering Provider Name)
                    -- Loop 2330E (Other Payer Service Facility Location)
                    -- Loop 2330F (Other Payer Supervising Provider Name)
                    -- Loop 2330G (Other Payer Billing Provider Name)
                -- Loop 2400 (Service Line)
                    -- Loop 2410 (Drug Identification Name)
                    -- Loop 2420A (Rendering Provider Name)
                    -- Loop 2420B (Purchased Service Provider Name)
                    -- Loop 2420C (Service Facility Location Name)
                    -- Loop 2420D (Supervising Provider Name)
                    -- Loop 2420E (Ordering Provider Name)
                    -- Loop 2420F (Referring Provider Name)
                    -- Loop 2420G (Ambulance Pickup Location)
                    -- Loop 2420H (Ambulance Dropoff Location)
                    -- Loop 2430 (Line Adjudication Information)
                    -- Loop 2440 (Form Identification)
-- Footer

The Header and Footer components are not "loops" per the specification, but are included to standardize and simplify
transactional modeling and processing.
"""
from decimal import Decimal

from x12.models import X12SegmentGroup
from .segments import (
    HeaderStSegment,
    HeaderBhtSegment,
    Loop1000ANm1Segment,
    Loop1000APerSegment,
    Loop1000BNm1Segment,
    Loop2000AHlSegment,
    Loop2000APrvSegment,
    Loop2010AaNm1Segment,
    Loop2010AaRefSegment,
    Loop2010AbNm1Segment,
    Loop2010AcNm1Segment,
    Loop2000BHlSegment,
    Loop2000BSbrSegment,
    Loop2010BaNm1Segment,
    Loop2010BaPerSegment,
    Loop2010BbNm1Segment,
    Loop2010BbRefSegment,
    Loop2300DtpSegment,
    Loop2300PwkSegment,
    Loop2300Cn1Segment,
    Loop2300AmtSegment,
    Loop2300RefSegment,
    Loop2300NteSegment,
    Loop2300CrcSegment,
    Loop2000CHlSegment,
    Loop2000CPatSegment,
    Loop2010CaNm1Segment,
    Loop2010CaRefSegment,
    Loop2010CaPerSegment,
    Loop2310ANm1Segment,
    Loop2310ARefSegment,
    Loop2310BNm1Segment,
    Loop2310BPrvSegment,
    Loop2310BRefSegment,
    Loop2310CNm1Segment,
    Loop2310CRefSegment,
    Loop2310CPerSegment,
    Loop2310DNm1Segment,
    Loop2310DRefSegment,
    Loop2310ENm1Segment,
    Loop2310FNm1Segment,
    Loop2320SbrSegment,
    Loop2330aNm1Segment,
    Loop2330aRefSegment,
    Loop2330bNm1Segment,
    Loop2330BDtpSegment,
    Loop2300BRefSegment,
    Loop2330cNm1Segment,
    Loop2330cRefSegment,
    Loop2330dNm1Segment,
    Loop2330dRefSegment,
    Loop2330eNm1Segment,
    Loop2330eRefSegment,
    Loop2330fNm1Segment,
    Loop2330fRefSegment,
    Loop2330gNm1Segment,
    Loop2330gRefSegment,
    Loop2400PwkSegment,
    Loop2400CrcSegment,
    Loop2400DtpSegment,
    Loop2400QtySegment,
    Loop2400Cn1Segment,
    Loop2400RefSegment,
    Loop2400AmtSegment,
    Loop2400NteSegment,
    Loop2410RefSegment,
    Loop2420ANm1Segment,
    Loop2420APrvSegment,
    Loop2420ARefSegment,
    Loop2420BNm1Segment,
    Loop2420BRefSegment,
    Loop2420CNm1Segment,
    Loop2420CRefSegment,
    Loop2420DNm1Segment,
    Loop2420DRefSegment,
    Loop2420ENm1Segment,
    Loop2420ERefSegment,
    Loop2420EPerSegment,
    Loop2420FNm1Segment,
    Loop2420FRefSegment,
    Loop2420GNm1Segment,
    Loop2420HNm1Segment,
    Loop2430DtpSegment,
    Loop2430AmtSegment,
    Loop2010AaPerSegment,
)
from x12.segments import (
    SeSegment,
    CurSegment,
    N3Segment,
    N4Segment,
    PatSegment,
    DmgSegment,
    RefSegment,
    ClmSegment,
    K3Segment,
    Cr1Segment,
    Cr2Segment,
    HiSegment,
    CasSegment,
    AmtSegment,
    OiSegment,
    MoaSegment,
    LxSegment,
    Sv1Segment,
    Sv5Segment,
    Cr3Segment,
    MeaSegment,
    Ps1Segment,
    HcpSegment,
    LinSegment,
    CtpSegment,
    SvdSegment,
    LqSegment,
    FrmSegment,
)
from typing import List, Optional, Dict
from pydantic import Field, root_validator
from x12.validators import (
    validate_duplicate_date_qualifiers,
    validate_duplicate_amt_codes,
)


class Loop1000A(X12SegmentGroup):
    """
    Loop 1000A - Submitter Name
    """

    nm1_segment: Loop1000ANm1Segment
    per_segment: List[Loop1000APerSegment] = Field(min_items=1, max_items=2)


class Loop1000B(X12SegmentGroup):
    """
    Loop 1000B - Receiver Name
    """

    nm1_segment: Loop1000BNm1Segment


class Loop2010Aa(X12SegmentGroup):
    """
    Loop 2010AA - Billing Provider Name
    """

    nm1_segment: Loop2010AaNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment
    ref_segment: List[Loop2010AaRefSegment] = Field(min_items=1, max_items=3)
    per_segment: Optional[List[Loop2010AaPerSegment]] = Field(min_items=0, max_items=2)


class Loop2010Ab(X12SegmentGroup):
    """
    Loop 2010AB - Pay to Address
    """

    nm1_segment: Loop2010AbNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment


class Loop2010Ac(X12SegmentGroup):
    """
    Loop 2010AC - Pay to Plan
    """

    nm1_segment: Loop2010AcNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment
    ref_segment: List[Loop2010AaRefSegment] = Field(min_items=1, max_items=2)


class Loop2010Ba(X12SegmentGroup):
    """
    Loop 2010BA - Subscriber Name
    """

    nm1_segment: Loop2010BaNm1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    dmg_segment: Optional[DmgSegment]
    ref_segment: List[Optional[RefSegment]] = Field(min_items=0, max_items=2)
    per_segment: Optional[Loop2010BaPerSegment]


class Loop2010Bb(X12SegmentGroup):
    """
    Loop 2010Bb - Payer Name
    """

    nm1_segment: Loop2010BbNm1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    ref_segment: Optional[List[Loop2010BbRefSegment]]


class Loop2330G(X12SegmentGroup):
    """
    Claim - Other Subscriber Other Payer Billing Provider
    """

    nm1_segment: Loop2330gNm1Segment
    ref_segment: List[Loop2330gRefSegment] = Field(min_items=1, max_items=3)


class Loop2330F(X12SegmentGroup):
    """
    Claim - Other Subscriber Other Payer Supervising Provider
    """

    nm1_segment: Loop2330fNm1Segment
    ref_segment: List[Loop2330fRefSegment] = Field(min_items=1, max_items=3)


class Loop2330E(X12SegmentGroup):
    """
    Claim - Other Subscriber Other Payer Service Facility Location
    """

    nm1_segment: Loop2330eNm1Segment
    ref_segment: List[Loop2330eRefSegment] = Field(min_items=1, max_items=3)


class Loop2330D(X12SegmentGroup):
    """
    Claim - Other Subscriber Other Payer Rendering Provider
    """

    nm1_segment: Loop2330dNm1Segment
    ref_segment: List[Loop2330dRefSegment] = Field(min_items=1, max_items=3)


class Loop2330C(X12SegmentGroup):
    """
    Claim - Other Subscriber Other Payer Referring Provider
    """

    nm1_segment: Loop2330cNm1Segment
    ref_segment: List[Loop2330cRefSegment] = Field(min_items=1, max_items=3)


class Loop2330B(X12SegmentGroup):
    """
    Claim - Other Payer Name
    """

    nm1_segment: Loop2330bNm1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    dtp_segment: Optional[Loop2330BDtpSegment]
    ref_segment: Optional[List[Loop2300BRefSegment]] = Field(min_items=0, max_items=6)


class Loop2330A(X12SegmentGroup):
    """
    Claim - Other Subscriber Name
    """

    nm1_segment: Loop2330aNm1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    ref_segment: Optional[Loop2330aRefSegment]


class Loop2320(X12SegmentGroup):
    """
    Claim Other subscriber information
    """

    sbr_segment: Loop2320SbrSegment
    cas_segment: Optional[List[CasSegment]] = Field(min_items=0, max_items=5)
    amt_segment: Optional[List[AmtSegment]] = Field(min_items=0, max_items=3)
    oi_segment: Optional[OiSegment]
    moa_segment: Optional[MoaSegment]
    loop_2330a: Optional[Loop2330A]
    loop_2330b: Optional[Loop2330B]
    loop_2330c: Optional[List[Loop2330C]]
    loop_2330d: Optional[Loop2330D]
    loop_2330e: Optional[Loop2330E]
    loop_2330f: Optional[Loop2330F]
    loop_2330g: Optional[Loop2330G]

    _validate_amt_segments = root_validator(allow_reuse=True)(
        validate_duplicate_amt_codes
    )


class Loop2310F(X12SegmentGroup):
    """
    Claim ambulance drop-off location
    """

    nm1_segment: Loop2310FNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment


class Loop2310E(X12SegmentGroup):
    """
    Claim ambulance pickup location
    """

    nm1_segment: Loop2310ENm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment


class Loop2310D(X12SegmentGroup):
    """
    Claim Supervising Provider
    """

    nm1_segment: Loop2310DNm1Segment
    ref_segment: Optional[List[Loop2310DRefSegment]]


class Loop2310C(X12SegmentGroup):
    """
    Claim Service Facility Location
    """

    nm1_segment: Optional[Loop2310CNm1Segment]
    n3_segment: N3Segment
    n4_segment: N4Segment
    ref_segment: Optional[List[Loop2310CRefSegment]] = Field(min_items=0, max_items=3)
    per_segment: Optional[Loop2310CPerSegment]


class Loop2310B(X12SegmentGroup):
    """
    Claim Rendering Provider
    """

    nm1_segment: Loop2310BNm1Segment
    prv_segment: Optional[Loop2310BPrvSegment]
    ref_segment: Optional[List[Loop2310BRefSegment]] = Field(min_items=0, max_items=4)


class Loop2310A(X12SegmentGroup):
    """
    Claim Referring Provider
    """

    nm1_segment: Loop2310ANm1Segment
    ref_segment: Optional[List[Loop2310ARefSegment]] = Field(min_items=0, max_items=3)


class Loop2440(X12SegmentGroup):
    """
    Form Identification Code
    """

    lq_segment: LqSegment
    frm_segment: List[FrmSegment] = Field(min_items=1, max_items=99)


class Loop2430(X12SegmentGroup):
    """
    Claim Service Line - Adjudication Information
    """

    svd_segment: SvdSegment
    cas_segment: Optional[List[CasSegment]] = Field()
    dtp_segment: Loop2430DtpSegment
    amt_segment: Optional[Loop2430AmtSegment]


class Loop2420H(X12SegmentGroup):
    """
    Claim Service Line - Ambulance Drop-Off Location
    """

    nm1_segment: Loop2420HNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment


class Loop2420G(X12SegmentGroup):
    """
    Claim Service Line - Ambulance Pick-Up Location
    """

    nm1_segment: Loop2420GNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment


class Loop2420F(X12SegmentGroup):
    """
    Claim Service Line - Referring Provider Name
    """

    nm1_segment: Loop2420FNm1Segment
    ref_segment: Optional[List[Loop2420FRefSegment]] = Field(min_items=0, max_items=20)


class Loop2420E(X12SegmentGroup):
    """
    Claim Service Line - Ordering Provider Name
    """

    nm1_segment: Loop2420ENm1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    ref_segment: Optional[List[Loop2420ERefSegment]] = Field(min_items=0, max_items=20)
    per_segment: Optional[Loop2420EPerSegment]


class Loop2420D(X12SegmentGroup):
    """
    Claim Service Line - Supervising Provider Name
    """

    nm1_segment: Loop2420DNm1Segment
    ref_segment: Optional[List[Loop2420DRefSegment]] = Field(min_items=0, max_items=20)


class Loop2420C(X12SegmentGroup):
    """
    Claim Service Line - Service Facility Location Name
    """

    nm1_segment: Loop2420CNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment
    ref_segment: Optional[List[Loop2420CRefSegment]] = Field(min_items=0, max_items=3)


class Loop2420B(X12SegmentGroup):
    """
    Claim Service Line - Purchased Service Provider
    """

    nm1_segment: Loop2420BNm1Segment
    ref_segment: Optional[List[Loop2420BRefSegment]] = Field(min_items=0, max_items=20)


class Loop2420A(X12SegmentGroup):
    """
    Claim Service Line - Rendering Provider
    """

    nm1_segment: Loop2420ANm1Segment
    prv_segment: Optional[Loop2420APrvSegment]
    ref_segment: Optional[List[Loop2420ARefSegment]] = Field(min_items=0, max_items=20)


class Loop2410(X12SegmentGroup):
    """
    Claim Service Line - Drug Identification
    """

    lin_segment: LinSegment
    ctp_segment: CtpSegment
    ref_segment: Optional[Loop2410RefSegment]


class Loop2400(X12SegmentGroup):
    """
    Claim - Service Line
    """

    lx_segment: LxSegment
    sv1_segment: Sv1Segment
    sv5_segment: Optional[Sv5Segment]
    pwk_segment: Optional[List[Loop2400PwkSegment]] = Field(min_items=0, max_items=10)
    cr1_segment: Optional[Cr1Segment]
    cr3_segment: Optional[Cr3Segment]
    crc_segment: Optional[List[Loop2400CrcSegment]] = Field(min_items=0, max_items=5)
    dtp_segment: List[Loop2400DtpSegment] = Field(min_items=1, max_items=11)
    qty_segment: Optional[List[Loop2400QtySegment]] = Field(min_items=0, max_items=2)
    mea_segment: Optional[List[MeaSegment]] = Field(min_items=0, max_items=5)
    cn1_segment: Optional[Loop2400Cn1Segment]
    ref_segment: Optional[List[Loop2400RefSegment]] = Field(min_items=0, max_items=11)
    amt_segment: Optional[List[Loop2400AmtSegment]] = Field(min_items=0, max_items=2)
    k3_segment: Optional[List[K3Segment]] = Field(min_items=0, max_items=10)
    nte_segment: Optional[List[Loop2400NteSegment]] = Field(min_items=0, max_items=2)
    ps1_segment: Optional[Ps1Segment]
    hcp_segment: Optional[HcpSegment]
    loop_2410: Optional[Loop2410]
    loop_2420a: Optional[Loop2420A]
    loop_2420b: Optional[Loop2420B]
    loop_2420c: Optional[Loop2420C]
    loop_2420d: Optional[Loop2420D]
    loop_2420e: Optional[Loop2420E]
    loop_2420f: Optional[Loop2420F]
    loop_2420g: Optional[Loop2420G]
    loop_2420h: Optional[Loop2420H]
    loop_2430: Optional[List[Loop2430]] = Field(min_items=0, max_items=15)
    loop_2440: Optional[List[Loop2440]] = Field(min_items=0)

    _validate_dtp_qualifiers = root_validator(allow_reuse=True)(
        validate_duplicate_date_qualifiers
    )


class Loop2300(X12SegmentGroup):
    """
    Loop 2300 - Claims
    """

    clm_segment: ClmSegment
    dtp_segment: Optional[List[Loop2300DtpSegment]] = Field(min_items=0, max_items=17)
    pwk_segment: Optional[List[Loop2300PwkSegment]] = Field(min_items=0, max_items=10)
    cn1_segment: Optional[Loop2300Cn1Segment]
    amt_segment: Optional[Loop2300AmtSegment]
    ref_segment: Optional[List[Loop2300RefSegment]] = Field(min_items=0, max_items=14)
    k3_segment: Optional[List[K3Segment]] = Field(min_items=0, max_items=10)
    nte_segment: Optional[Loop2300NteSegment]
    cr1_segment: Optional[Cr1Segment]
    cr2_segment: Optional[Cr2Segment]
    crc_segment: Optional[List[Loop2300CrcSegment]] = Field(min_items=0, max_items=8)
    hi_segment: Optional[List[HiSegment]] = Field(min_items=1, max_items=4)
    hcp_segment: Optional[HcpSegment]
    loop_2310a: Optional[Loop2310A]
    loop_2310b: Optional[Loop2310B]
    loop_2310c: Optional[Loop2310C]
    loop_2310d: Optional[Loop2310D]
    loop_2310e: Optional[Loop2310E]
    loop_2310f: Optional[Loop2310F]
    loop_2320: Optional[List[Loop2320]] = Field(min_items=0, max_items=10)
    loop_2400: List[Loop2400] = Field(min_items=1, max_items=50)

    _validate_dtp_qualifiers = root_validator(allow_reuse=True)(
        validate_duplicate_date_qualifiers
    )

    @root_validator
    def validate_claim_amounts(cls, values: Dict):
        """
        Validates that CLM02 == SUM(Loop2400.SV102)
        """
        claim_amount: Decimal = values.get("clm_segment").total_claim_charge_amount
        line_total: Decimal = Decimal("0.0")

        for line in values.get("loop_2400", []):
            line_total += line.sv1_segment.line_item_charge_amount

        if claim_amount != line_total:
            raise ValueError(
                f"Claim Amount {claim_amount} != Service Line Total {line_total}"
            )

        return values


class Loop2010Ca(X12SegmentGroup):
    """
    Loop 2010CA Patient Name
    """

    nm1_segment: Loop2010CaNm1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment
    dmg_segment: DmgSegment
    ref_segment: Optional[List[Loop2010CaRefSegment]] = Field(min_items=0, max_items=2)
    per_segment: Optional[Loop2010CaPerSegment]


class Loop2000C(X12SegmentGroup):
    """
    Loop 2000C - Patient
    """

    hl_segment: Loop2000CHlSegment
    pat_segment: Loop2000CPatSegment
    loop_2010ca: Loop2010Ca
    loop_2300: Optional[List[Loop2300]] = Field(min_items=0, max_items=100)


class Loop2000B(X12SegmentGroup):
    """
    Loop 2000B - Subscriber
    """

    hl_segment: Loop2000BHlSegment
    sbr_segment: Loop2000BSbrSegment
    pat_segment: Optional[PatSegment]
    loop_2010ba: Loop2010Ba
    loop_2010bb: Loop2010Bb
    loop_2300: Optional[List[Loop2300]] = Field(min_items=0, max_items=100)
    loop_2000c: Optional[List[Loop2000C]]


class Loop2000A(X12SegmentGroup):
    """
    Loop 2000A - Billing Provider
    """

    hl_segment: Loop2000AHlSegment
    prv_segment: Optional[Loop2000APrvSegment]
    cur_segment: Optional[CurSegment]
    loop_2010aa: Loop2010Aa
    loop_2010ab: Optional[Loop2010Ab]
    loop_2010ac: Optional[Loop2010Ac]
    loop_2000b: List[Loop2000B]


class Header(X12SegmentGroup):
    """
    Transaction Header Information
    """

    st_segment: HeaderStSegment
    bht_segment: HeaderBhtSegment


class Footer(X12SegmentGroup):
    """
    Transaction Footer Information
    """

    se_segment: SeSegment



"""
parsing.py

Parses X12 271 005010X279A1 segments into a transactional domain model.

The parsing module includes a specific parser for the 271 transaction and loop parsing functions to create new loops
as segments are streamed to the transactional data model.

Loop parsing functions are implemented as set_[description]_loop(context: X12ParserContext, segment_data: Dict).
"""
from enum import Enum

from x12.parsing import match, X12ParserContext
from typing import Dict


class TransactionLoops(str, Enum):
    """
    The loops used to support the 271 005010X279A1 format.
    """

    HEADER = "header"
    INFORMATION_SOURCE = "loop_2000a"
    INFORMATION_SOURCE_NAME = "loop_2100a"
    INFORMATION_RECEIVER = "loop_2000b"
    INFORMATION_RECEIVER_NAME = "loop_2100b"
    SUBSCRIBER = "loop_2000c"
    SUBSCRIBER_NAME = "loop_2100c"
    SUBSCRIBER_ELIGIBILITY = "loop_2110c"
    SUBSCRIBER_ADDITIONAL_ELIGIBILITY = "loop_2115c"
    SUBSCRIBER_RELATED_BENEFIT_NAME = "loop_2120c"
    DEPENDENT = "loop_2000d"
    DEPENDENT_NAME = "loop_2100d"
    DEPENDENT_ELIGIBILITY = "loop_2110d"
    DEPENDENT_ADDITIONAL_ELIGIBILITY = "loop_2115d"
    DEPENDENT_RELATED_BENEFIT_NAME = "loop_2120d"
    FOOTER = "footer"


# Subscriber HL Segment Hierarchy Level Code
SUBSCRIBER_HIERARCHY_LEVEL = "22"


def _get_info_source(context) -> Dict:
    """Returns the current information source"""
    return context.transaction_data[TransactionLoops.INFORMATION_SOURCE][-1]


def _get_info_receiver(context) -> Dict:
    """Returns the current information receiver"""
    info_source = _get_info_source(context)
    return info_source[TransactionLoops.INFORMATION_RECEIVER][-1]


def _is_subscriber_patient(context) -> bool:
    """Returns true if the current context contains a subscriber that is the patient"""
    hl_level_code = context.hl_segment.get("hierarchical_level_code")
    hl_child_code = context.hl_segment.get("hierarchical_child_code")
    return hl_level_code == SUBSCRIBER_HIERARCHY_LEVEL and hl_child_code == "0"


def _get_eligibility(context) -> Dict:
    """Returns the current eligibility record"""
    if _is_subscriber_patient(context):
        name_loop = TransactionLoops.SUBSCRIBER_NAME
        eligibility_loop = TransactionLoops.SUBSCRIBER_ELIGIBILITY
    else:
        name_loop = TransactionLoops.DEPENDENT_NAME
        eligibility_loop = TransactionLoops.DEPENDENT_ELIGIBILITY

    return context.patient_record[name_loop][eligibility_loop][-1]


@match("ST")
def set_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set header loop.
    Creates initial data structures for the record

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    context.transaction_data[TransactionLoops.HEADER] = {}
    context.set_loop_context(
        TransactionLoops.HEADER, context.transaction_data[TransactionLoops.HEADER]
    )


@match("HL", {"hierarchical_level_code": "20"})
def set_information_source_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the information source loop (Loop 2000A)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if TransactionLoops.INFORMATION_SOURCE not in context.transaction_data:
        context.transaction_data[TransactionLoops.INFORMATION_SOURCE] = []

    context.transaction_data[TransactionLoops.INFORMATION_SOURCE].append(
        {"aaa_segment": []}
    )
    information_source = context.transaction_data[TransactionLoops.INFORMATION_SOURCE][
        -1
    ]
    context.set_loop_context(TransactionLoops.INFORMATION_SOURCE, information_source)


@match("NM1")
def set_information_source_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the information source name (Loop 2100A)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name == TransactionLoops.INFORMATION_SOURCE:
        context.loop_container[TransactionLoops.INFORMATION_SOURCE_NAME] = {
            "per_segment": [],
            "aaa_segment": [],
        }
        context.set_loop_context(
            TransactionLoops.INFORMATION_SOURCE_NAME,
            context.loop_container[TransactionLoops.INFORMATION_SOURCE_NAME],
        )


@match("HL", {"hierarchical_level_code": "21"})
def set_information_receiver_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the information receiver loop (Loop 2000B)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    info_source = _get_info_source(context)

    if TransactionLoops.INFORMATION_RECEIVER not in info_source:
        info_source[TransactionLoops.INFORMATION_RECEIVER] = []

    info_source[TransactionLoops.INFORMATION_RECEIVER].append({})
    info_receiver = info_source[TransactionLoops.INFORMATION_RECEIVER][-1]
    context.set_loop_context(TransactionLoops.INFORMATION_RECEIVER, info_receiver)


@match("NM1")
def set_information_receiver_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the information receiver name (Loop 2100B)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name == TransactionLoops.INFORMATION_RECEIVER:
        context.loop_container[TransactionLoops.INFORMATION_RECEIVER_NAME] = {
            "ref_segment": [],
            "aaa_segment": [],
        }
        context.set_loop_context(
            TransactionLoops.INFORMATION_RECEIVER_NAME,
            context.loop_container[TransactionLoops.INFORMATION_RECEIVER_NAME],
        )


@match("HL", {"hierarchical_level_code": "22"})
def set_subscriber_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber loop (Loop 2000C)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    info_receiver = _get_info_receiver(context)

    if TransactionLoops.SUBSCRIBER not in info_receiver:
        info_receiver[TransactionLoops.SUBSCRIBER] = []

    info_receiver[TransactionLoops.SUBSCRIBER].append({"trn_segment": []})
    context.subscriber_record = info_receiver[TransactionLoops.SUBSCRIBER][-1]
    context.set_loop_context(TransactionLoops.SUBSCRIBER, context.subscriber_record)

    if _is_subscriber_patient(context):
        context.patient_record = context.subscriber_record


@match("NM1", conditions={"entity_identifier_code": "IL"})
def set_subscriber_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber loop (Loop 2100C)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name == TransactionLoops.SUBSCRIBER:
        context.subscriber_record[TransactionLoops.SUBSCRIBER_NAME] = {
            "ref_segment": [],
            "aaa_segment": [],
            "dtp_segment": [],
        }
        context.set_loop_context(
            TransactionLoops.SUBSCRIBER_NAME,
            context.subscriber_record[TransactionLoops.SUBSCRIBER_NAME],
        )


@match("EB")
def set_eligibility_benefit_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the eligibility benefit loop for subscriber (2110C) or dependents (2110D)
    """
    if _is_subscriber_patient(context):
        name_loop = TransactionLoops.SUBSCRIBER_NAME
        eligibility_loop = TransactionLoops.SUBSCRIBER_ELIGIBILITY
    else:
        name_loop = TransactionLoops.DEPENDENT_NAME
        eligibility_loop = TransactionLoops.DEPENDENT_ELIGIBILITY

    if eligibility_loop not in context.patient_record[name_loop]:
        context.patient_record[name_loop][eligibility_loop] = []

    context.patient_record[name_loop][eligibility_loop].append(
        {
            "hsd_segment": [],
            "ref_segment": [],
            "dtp_segment": [],
            "aaa_segment": [],
            "msg_segment": [],
        }
    )
    eligibility_record = context.patient_record[name_loop][eligibility_loop][-1]
    context.set_loop_context(eligibility_loop, eligibility_record)


@match("III")
def set_additional_benefit_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the additional benefit information loop - subscriber (Loop 2115C), dependent (Loop 2115D)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if _is_subscriber_patient(context):
        additional_benefit_loop = TransactionLoops.SUBSCRIBER_ADDITIONAL_ELIGIBILITY
    else:
        additional_benefit_loop = TransactionLoops.DEPENDENT_ADDITIONAL_ELIGIBILITY

    eligibility = _get_eligibility(context)

    if additional_benefit_loop not in eligibility:
        eligibility[additional_benefit_loop] = []

    eligibility[additional_benefit_loop].append({})
    additional_benefit = eligibility[additional_benefit_loop][-1]
    context.set_loop_context(additional_benefit_loop, additional_benefit)


@match("LS")
def set_loop_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Resets the loop context to the Eligibility Loop 2120C/D (Subscriber and Dependent)
    The LS segment precedes the Subscriber/Dependent Benefit Related Entity Name

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    eligibility = _get_eligibility(context)

    if _is_subscriber_patient(context):
        loop_name = TransactionLoops.SUBSCRIBER_ELIGIBILITY
    else:
        loop_name = TransactionLoops.DEPENDENT_ELIGIBILITY

    context.set_loop_context(loop_name, eligibility)


@match("NM1")
def set_benefit_related_entity_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Supports the Benefit Related Entity Name Loop for the Subscriber (2120C) and Dependent (2120D).

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name not in (
        TransactionLoops.DEPENDENT_ELIGIBILITY,
        TransactionLoops.SUBSCRIBER_ELIGIBILITY,
    ):
        return

    if _is_subscriber_patient(context):
        benefit_related_entity_loop = TransactionLoops.SUBSCRIBER_RELATED_BENEFIT_NAME
    else:
        benefit_related_entity_loop = TransactionLoops.DEPENDENT_RELATED_BENEFIT_NAME

    eligibility = _get_eligibility(context)

    if benefit_related_entity_loop not in eligibility:
        eligibility[benefit_related_entity_loop] = []

    eligibility[benefit_related_entity_loop].append(
        {
            "per_segment": [],
        }
    )

    benefit_related_entity = eligibility[benefit_related_entity_loop][-1]
    context.set_loop_context(benefit_related_entity_loop, benefit_related_entity)


@match("LE")
def set_loop_footer_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Ensures that the Loop Footer Segment is written to the Member (Subscriber or Dependent)
    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    eligibility = _get_eligibility(context)

    if _is_subscriber_patient(context):
        context.set_loop_context(TransactionLoops.SUBSCRIBER_ELIGIBILITY, eligibility)
    else:
        context.set_loop_context(TransactionLoops.DEPENDENT_ELIGIBILITY, eligibility)


@match("HL", {"hierarchical_level_code": "23"})
def set_dependent_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the dependent loop (Loop 2000D)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if TransactionLoops.DEPENDENT not in context.subscriber_record:
        context.subscriber_record[TransactionLoops.DEPENDENT] = []

    context.subscriber_record[TransactionLoops.DEPENDENT].append({"trn_segment": []})
    context.patient_record = context.subscriber_record[TransactionLoops.DEPENDENT][-1]
    context.set_loop_context(TransactionLoops.DEPENDENT, context.patient_record)


@match("NM1", conditions={"entity_identifier_code": "03"})
def set_dependent_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the dependent name loop (Loop 2100D)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    if context.loop_name == TransactionLoops.DEPENDENT:
        context.patient_record[TransactionLoops.DEPENDENT_NAME] = {
            "ref_segment": [],
            "aaa_segment": [],
            "dtp_segment": [],
        }
        context.set_loop_context(
            TransactionLoops.DEPENDENT_NAME,
            context.patient_record[TransactionLoops.DEPENDENT_NAME],
        )


@match("SE")
def set_se_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )



"""
segments.py

Specialized segment models for the HealthCare Claim Institutional 837 005010X223A3 transaction.
"""

from x12.segments import (
    StSegment,
    BhtSegment,
    Nm1Segment,
    PerSegment,
    HlSegment,
    PrvSegment,
    RefSegment,
    SbrSegment,
    DtpSegment,
    Cn1Segment,
    AmtSegment,
    NteSegment,
    CrcSegment,
    PatSegment,
    HcpSegment,
    PwkSegment,
)
from typing import Literal, Optional, Dict, Union
from enum import Enum
from pydantic import Field, root_validator
import datetime


class HeaderStSegment(StSegment):
    """
    Customized ST segment for 837 transaction header
    """

    transaction_set_identifier_code: Literal["837"]
    implementation_convention_reference: Literal["005010X223A3"]


class HeaderBhtSegment(BhtSegment):
    """
    Customized BHT segment for 837 transaction header
    """

    class PurposeCode(str, Enum):
        """
        BHT02 code values
        """

        ORIGINAL = "00"
        REISSUE = "18"

    class TransactionTypeCode(str, Enum):
        """
        BHT06 code values
        """

        SUBROGATION_DEMAND = "31"
        CHARGABLE = "CH"
        REPORTING = "RP"

    hierarchical_structure_code: Literal["0019"]
    transaction_set_purpose_code: PurposeCode
    transaction_type_code: TransactionTypeCode


class Loop1000ANm1Segment(Nm1Segment):
    """
    Submitter Name Name Segment
    """

    entity_identifier_code: Literal["41"]
    identification_code_qualifier: Literal["46"]
    identification_code: str = Field(min_length=2, max_length=80)


class Loop1000APerSegment(PerSegment):
    """
    Submitter Name Contact Segment
    """

    class CommunicationNumberQualifier1(str, Enum):
        """
        Code values for PER03
        """

        ELECTRONIC_MAIL = "EM"
        FACSIMILE = "FX"
        TELEPHONE = "TE"

    class CommunicationNumberQualifier2(str, Enum):
        """
        Code values for PER05
        """

        ELECTRONIC_MAIL = "EM"
        TELEPHONE_EXTENSION = "EX"
        FACSIMILE = "FX"
        TELEPHONE = "TE"

    communication_number_qualifier_1: CommunicationNumberQualifier1
    communication_number_qualifier_2: Optional[CommunicationNumberQualifier2]
    communication_number_qualifier_3: Optional[CommunicationNumberQualifier2]


class Loop1000BNm1Segment(Nm1Segment):
    """
    Information Receiver Name
    """

    entity_identifier_code: Literal["40"]
    entity_type_qualifier: Literal["2"]
    identification_code_qualifier: Literal["46"]


class Loop2000AHlSegment(HlSegment):
    """
    Billing Provider Hierarchical Level
    """

    hierarchical_parent_id_number: Optional[str]
    hierarchical_level_code: Literal["20"]
    hierarchical_child_code: Literal["1"]


class Loop2000APrvSegment(PrvSegment):
    """
    Billing Provider Speciality Information
    """

    provider_code: Literal["BI"]
    reference_identification_qualifier: Literal["PXC"]
    reference_identification: str = Field(min_length=1, max_length=50)


class Loop2000BHlSegment(HlSegment):
    """
    Subscriber Hierarchy Segment
    """

    hierarchical_level_code: Literal["22"]


class Loop2000BSbrSegment(SbrSegment):
    """
    Subscriber Information Segment

    TODO: validation rules
    """

    class PayerResponsibilityCode(str, Enum):
        """
        SBR01 code values
        """

        PAYER_RESPONSIBILITY_FOUR = "A"
        PAYER_RESPONSIBILITY_FIVE = "B"
        PAYER_RESPONSIBILITY_SIX = "C"
        PAYER_RESPONSIBILITY_SEVEN = "D"
        PAYER_RESPONSIBILITY_EIGHT = "E"
        PAYER_RESPONSIBILITY_NINE = "F"
        PAYER_RESPONSIBILITY_TEN = "G"
        PAYER_RESPONSIBILITY_ELEVEN = "H"
        PRIMARY = "P"
        SECONDARY = "S"
        TERTIARY = "T"
        UNKNOWN = "U"

    class InsuranceTypeCode(str, Enum):
        """
        SBR05 code values
        """

        MEDICARE_SECONDARY_WORKING = "12"
        MEDICARE_SECONDARY_END_STAGE_RENAL = "13"
        MEDICARE_SECONDARY_AUTO_PRIMARY = "14"
        MEDICARE_SECONDARY_WORKERS_COMPENSATION = "15"
        MEDICARE_SECONDARY_PUBLIC_HEALTH_SERVICE = "16"
        MEDICARE_SECONDARY_BLACK_LUNG = "41"
        MEDICARE_SECONDARY_VETERANS_ADMINISTRATION = "42"
        MEDICARE_SECONDARY_DISABLED_BENEFICIARY = "43"
        MEDICARE_SECONDARY_LIABILITY_PRIMARY = "47"

    class ClaimFilingIndicatorCode(str, Enum):
        """
        SBR09 code values
        """

        OTHER_NON_FEDERAL_PROGRAMS = "11"
        PREFERRED_PROVIDER_ORGANIZATION = "12"
        POINT_OF_SERVICE = "13"
        EXCLUSIVE_PROVIDER_ORGANIZATION = "14"
        INDEMNITY_INSURANCE = "15"
        HEALTH_MAINTENANCE_ORGANIZATION_MEDICARE_RISK = "16"
        DENTAL_MAINTENANCE_ORGANIZATION = "17"
        AUTOMOBILE_MEDICAL = "AM"
        BLUE_CROSS_BLUE_SHIELD = "BL"
        CHAMPUS = "CH"
        COMMERICIAL_INSURANCE_COMPANY = "CI"
        DISABILITY = "DS"
        FEDERAL_EMPLOYEES_PROGRAM = "FI"
        HEALTH_MAINTENANCE_ORGANIZATION = "HM"
        LIABILITY_MEDICAL = "LM"
        MEDICARE_PART_A = "MA"
        MEDICARE_PART_B = "MB"
        MEDICAID = "MC"
        OTHER_FEDERAL_PROGRAM = "OF"
        TITLE_V = "TV"
        VETERANS_AFFAIR_PLAN = "VA"
        WORKERS_COMPENSATION_HEALTH_CLAIM = "WC"
        MUTUALLY_DEFINED = "ZZ"

    payer_responsibility_code: PayerResponsibilityCode
    individual_relationship_code: Optional[Literal["18"]]
    insurance_type_code: Optional[InsuranceTypeCode]


class Loop2000CHlSegment(HlSegment):
    """
    The HL segment for Loop 2000C (Patient)
    """

    hierarchical_level_code: Literal["23"]
    hierarchical_child_code: Literal["0"]


class Loop2000CPatSegment(PatSegment):
    """
    The PAT, patient, segment for Loop2000C
    """

    class IndividualRelationshipCode(str, Enum):
        """
        Code values for PAT01
        """

        SPOUSE = "01"
        CHILD = "19"
        EMPLOYEE = "20"
        UNKNOWN = "21"
        ORGAN_DONOR = "39"
        CADAVER_DONOR = "40"
        LIFE_PARTNER = "53"
        OTHER_RELATIONSHIP = "G8"

    individual_relationship_code: IndividualRelationshipCode


class Loop2010AaNm1Segment(Nm1Segment):
    """
    Billing Provider Name and Identification
    """

    entity_identifier_code: Literal["85"]
    entity_type_qualifier: Literal["2"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2010AaRefSegment(RefSegment):
    """
    Billing Provider Identification Codes
    """

    reference_identification_qualifier: Literal["EI"]


class Loop2010AaPerSegment(PerSegment):
    """
    Billing Provider Contact Information
    """

    class CommunicationNumberQualifier(str, Enum):
        ELECTRONIC_MAIL = "EM"
        TELEPHONE_EXTENSION = "EX"
        FACSIMILE = "FX"
        TELEPHONE = "TE"

    communication_number_qualifier_1: CommunicationNumberQualifier
    communication_number_qualifier_2: Optional[CommunicationNumberQualifier]
    communication_number_qualifier_3: Optional[CommunicationNumberQualifier]


class Loop2010AbNm1Segment(Nm1Segment):
    """
    Billing Provider Name and Identification
    """

    entity_identifier_code: Literal["87"]
    entity_type_qualifier: Literal["2"]
    name_last_or_organization_name: Optional[str]


class Loop2010AcNm1Segment(Nm1Segment):
    """
    Pay to Plan Name and Identification
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        PAYOR_IDENTIFICATION = "PI"
        CMS_PLAN_ID = "XV"

    entity_identifier_code: Literal["PE"]
    entity_type_qualifier: Literal["2"]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop2010AcRefSegment(RefSegment):
    """
    Pay to Plan Identification Codes
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYER_IDENTIFICATION_NUMBER = "2U"
        CLAIM_OFFICE_NUMBER = "FY"
        NAIC_CODE = "NF"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2010BaNm1Segment(Nm1Segment):
    """
    Subscriber Name Segment
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        STANDARD_HEALTH_ID_US = "II"
        MEMBER_IDENTIFICATION_NUMBER = "MI"

    entity_identifier_code: Literal["IL"]
    identification_code_qualifier: Optional[IdentificationCodeQualifier]


class Loop2010BaRefSegment(RefSegment):
    """
    Subscriber Name Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        SOCIAL_SECURITY_NUMBER = "SY"
        AGENCY_CLAIM_NUMBER = "Y4"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2010BbNm1Segment(Nm1Segment):
    """
    Subscriber Name Segment
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        PAYOR_IDENTIFICATION = "PI"
        CMS_PLAN_ID = "XV"

    entity_identifier_code: Literal["PR"]
    entity_type_qualifier: Literal["2"]
    identification_code_qualifier: IdentificationCodeQualifier
    identification_code: str = Field(min_length=2, max_length=80)


class Loop2010BbRefSegment(RefSegment):
    """
    Payer Name Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYER_IDENTIFICATION_NUMBER = "2U"
        EMPLOYERS_IDENTIFICATION_NUMBER = "EI"
        CLAIM_OFFICER_NUMBER = "FY"
        NAIC_CODE = "NF"
        PROVIDER_CONTROL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2010CaNm1Segment(Nm1Segment):
    """
    Loop2010CA NM1 segment for Patient
    """

    entity_identifier_code: Literal["QC"]
    entity_type_qualifier: Literal["1"]


class Loop2010CaRefSegment(RefSegment):
    """
    Loop 2010CA (Patient Name) REF Segment
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        AGENCY_CLAIM_NUMBER = "Y4"
        MEMBER_IDENTIFICATION_NUMBER = "1W"
        SOCIAL_SECURITY_NUMBER = "SY"


class Loop2300DtpSegment(DtpSegment):
    """
    Claim information DTP segments
    """

    class DateTimeQualifier(str, Enum):
        """
        Code values for DTP01
        """

        DISCHARGE = "096"
        STATEMENT = "434"
        ADMISSION = "435"
        RECEIVED = "050"

    class DateTimePeriodFormatQualifier(str, Enum):
        """
        Code values for DTP02
        """

        TIME = "TM"
        DATE = "D8"
        DATE_AND_TIME = "DT"
        DATE_RANGE = "RD8"

    date_time_qualifier: DateTimeQualifier
    date_time_period_format_qualifier: DateTimePeriodFormatQualifier
    date_time_period: Union[str, datetime.datetime, datetime.date]

    @root_validator
    def validate_disability_dates(cls, values: Dict):
        """
        Validates that a date range qualifier is used for disability dates.

        :param values: The model's values
        :return: The model's values
        """
        date_qualifier = values.get("date_time_qualifier")
        period_qualifier = values.get("date_time_period_format_qualifier")

        if date_qualifier == "314" and period_qualifier != "RD8":
            raise ValueError(
                "RD8 Date Time Period is required for Disability Dates (314)"
            )
        return values


class Loop2300PwkSegment(PwkSegment):
    """
    Claim information paperwork segment
    """

    class AttachmentReportTypeCode(str, Enum):
        """
        PWK01 code values
        """

        REPORT_JUSTIFYING_TREATMENT_BEYOND_UTILIZATION_GUIDELINES = "03"
        DRUGS_ADMINISTERED = "04"
        TREATMENT_DIAGNOSIS = "05"
        INITIAL_ASSESSMENT = "06"
        FUNCTIONAL_GOALS = "07"
        PLAN_OF_TREATMENT = "08"
        PROGRESS_REPORT = "09"
        CONTINUED_TREATMENT = "10"
        CHEMICAL_ANALYSIS = "11"
        CERTIFIED_TEST_REPORT = "13"
        JUSTIFICATION_FOR_ADMISSION = "15"
        RECOVERY_PLAN = "21"
        ALLERGIES_SENSITIVITIES_DOCUMENT = "A3"
        AUTOPSY_REPORT = "A4"
        AMBULANCE_CERTIFICATION = "AM"
        ADMISSION_SUMMARY = "AS"
        PRESCRIPTION = "B2"
        PHYSICIAN_ORDER = "B3"
        REFERRAL_FORM = "B4"
        BENCHMARK_TESTING_RESULTS = "BR"
        BASELINE = "BS"
        BLANKET_TESTING_RESULTS = "BT"
        CHIROPRACTIC_JUSTIFICATION = "CB"
        CONSENT_FORM = "CK"
        CERTIFICATION = "CT"
        DRUG_PROFILE_DOCUMENT = "D2"
        DENTAL_MODELS = "DA"
        DURABLE_MEDICAL_EQUIPMENT_PRESCRIPTION = "DB"
        DIAGNOSTIC_REPORT = "DG"
        DISCHARGE_MONITORING_REPORT = "DJ"
        DISCHARGE_SUMMARY = "DS"
        EXPLANATION_OF_BENEFITS = "EB"
        HEALTH_CERTIFICATE = "HC"
        HEALTH_CLINIC_RECORDS = "HR"
        IMMUNIZATION_RECORD = "I5"
        STATE_SCHOOL_IMMUNIZATION_RECORDS = "IR"
        LABORATORY_RESULTS = "LA"
        MEDICAL_RECORD_ATTACHMENT = "M1"
        MODELS = "MT"
        NURSING_NOTES = "NN"
        OPERATIVE_NOTE = "OB"
        OXYGEN_CONTENT_AVERAGING_REPORT = "OC"
        ORDERS_AND_TREATMENTS_DOCUMENT = "OD"
        OBJECTIVE_PHYSICAL_EXAMINATION = "OE"
        OXYGEN_THERAPY_CERTIFICATION = "OX"
        SUPPORT_DATA_FOR_CLAIM = "OZ"
        PATHOLOGY_REPORT = "P4"
        PATIENT_MEDICAL_HISTORY_DOCUMENT = "P5"
        PARENTAL_OR_ENTERAL_CERTIFICATION = "PE"
        PHYSICAL_THERAPY_NOTES = "PN"
        PROSTHETICS_OR_ORTHOTIC_CERTIFICATION = "PO"
        PARAMEDICAL_RESULTS = "PQ"
        PHYSICIANS_REPORT = "PY"
        PHYSICAL_THERAPY_CERTIFICATION = "PZ"
        RADIOLOGY_FILMS = "RB"
        RADIOLOGY_REPORTS = "RR"
        REPORT_OF_TESTS_AND_ANALYSIS_REPORT = "RT"
        RENEWABLE_OXYGEN_CONTENT_AVERAGING_REPORT = "RX"
        SYMPTOMS_DOCUMENT = "SG"
        DEATH_NOTIFICATION = "V5"
        PHOTOGRAPHS = "XP"

    class AttachmentTransmissionCode(str, Enum):
        """
        Code values for PWK02
        """

        AVAILABLE_ON_REQUEST_PROVIDER_SITE = "AA"
        BY_MAIL = "BM"
        ELECTRONICALLY_ONLY = "EL"
        EMAIL = "EM"
        FILE_TRANSFER = "FT"
        BY_FAX = "FX"


class Loop2300Cn1Segment(Cn1Segment):
    """
    Claim information contract information
    """

    class ContractTypeCode(str, Enum):
        """
        Code values for CN101
        """

        DIAGNOSIS_RELATED_GROUP = "01"
        PER_DIEM = "02"
        VARIABLE_PER_DIEM = "03"
        FIAT = "04"
        CAPITATED = "05"
        PERCENT = "06"
        OTHER = "09"

    contract_type_code: ContractTypeCode


class Loop2300AmtSegment(AmtSegment):
    """
    Monetary Amount
    """

    amount_qualifier_code: Literal["F3"]


class Loop2300RefSegment(RefSegment):
    """
    Claim information reference identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        SPECIAL_PAYMENT_REFERENCE_NUMBER = "4N"
        REFERRAL_NUMBER = "9F"
        PRIOR_AUTHORIZATION_NUMBER = "G1"
        ORIGINAL_REFERENCE_NUMBER = "F8"
        REPRICED_CLAIM_REFERENCE_NUMBER = "9A"
        ADJUSTED_REPRICED_CLAIM_REFERENCE_NUMBER = "9C"
        QUALIFIED_PRODUCTS_LIST = "LX"
        CLAIM_NUMBER = "D9"
        LOCATION_NUMBER = "LU"
        MEDICAL_RECORD_IDENTIFICATION_NUMBER = "EA"
        PROJECT_CODE = "P4"
        PEER_REVIEW_ORG_APPROVAL_NUMBER = "G4"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2300NteSegment(NteSegment):
    """
    Claim information note segment
    """

    class NoteReferenceCode(str, Enum):
        """
        Code values for NTE01
        """

        ALLERGIES = "ALG"
        GOALS_REHAB_POTENTIAL_DISCHARGE_PLANS = "DCP"
        DIAGNOSIS_DESCRIPTION = "DGN"
        DURABLE_MEDICAL_EQUIPMENT_SUPPLIES = "DME"
        MEDICATIONS = "MED"
        NUTRITIONAL_REQUIREMENTS = "NTR"
        ORDERS_FOR_DISCIPLINES_AND_TREATMENTS = "ODT"
        FUNCTIONAL_LIMITATIONS_REASON_HOMEBOUND = "RHB"
        REASONS_PATIENT_LEAVES_HOME = "RLH"
        TIMES_REASONS_PATIENT_NOT_AT_HOME = "RNH"
        UNUSUAL_HOME_SOCIAL_ENV_OR_BOTH = "SET"
        SAFETY_MEASURES = "SFM"
        SUPPLEMENTARY_PLAN_OF_TREATMENT = "SPT"
        UPDATED_INFORMATION = "UPI"


class Loop2300CrcEpSdtRefferal(CrcSegment):
    """
    Claim Information - Homebound indicator
    """

    class YesNoResponseCode(str, Enum):
        """
        Code values for CRC02
        """

        NO = "N"
        YES = "Y"

    class ConditionsIndicator(str, Enum):
        """
        Code values for CRC03 - CRC07
        """

        AVAILABLE_NOT_USED = "AV"
        NOT_USED = "NU"
        UNDER_TREATMENT = "S2"
        NEW_SERVICES_REQUESTED = "ST"

    code_category: Literal["ZZ"]
    certification_condition_indicator: YesNoResponseCode
    condition_code_1: ConditionsIndicator
    condition_code_2: Optional[ConditionsIndicator]
    condition_code_3: Optional[ConditionsIndicator]
    condition_code_4: Optional[ConditionsIndicator]
    condition_code_5: Optional[ConditionsIndicator]


class Loop2300HcpSegment(HcpSegment):
    class PricingMethodologyCodes(str, Enum):
        """
        Code values for HCP01
        """

        ZERO_PRICING_NOT_COVERED = "00"
        PRICED_AS_BILLED_100_PERCENT = "01"
        PRICED_AT_STANDARD_FEE_SCHEDULE = "02"
        PRICED_AT_CONTRACTUAL_PERCENTAGE = "03"
        BUNDLED_PRICING = "04"
        PEER_REVIEW_PRICING = "05"
        PER_DIEM_PRICING = "06"
        FLAT_RATE_PRICING = "07"
        COMBINATION_PRICING = "08"
        MATERNITY_PRICING = "09"
        OTHER_PRICING = "10"
        LOWER_OF_COST = "11"
        RATIO_OF_COST = "12"
        COST_REIMBURSED = "13"
        ADJUSTMENT_PRICING = "14"

    class MeasurementCodes(str, Enum):
        """
        Code values for HCP11
        """

        DAYS = "DA"
        UNIT = "UN"

    pricing_methodology: PricingMethodologyCodes
    unit_basis_measurement_code: Optional[MeasurementCodes]


class Loop2310ANm1Segment(Nm1Segment):
    """
    Claim Attending Provider Name
    """

    entity_identifier_code: Literal["71"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2310APrvSegment(PrvSegment):
    """
    Attending Provider Speciality Information
    """

    provider_code: Literal["AT"]
    reference_identification_qualifier: Literal["PXC"]
    reference_identification: str = Field(min_length=1, max_length=50)


class Loop2310ARefSegment(RefSegment):
    """
    Claim Referring Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310BNm1Segment(Nm1Segment):
    """
    Claim Operating Physician Name
    """

    entity_identifier_code: Literal["72"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Literal["XX"]


class Loop2310BRefSegment(RefSegment):
    """
    Claim ROperating Physician Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310CNm1Segment(Nm1Segment):
    """
    Claim Other Operating Physician Name
    """

    entity_identifier_code: Literal["ZZ"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2310CRefSegment(RefSegment):
    """
    Claim Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310DNm1Segment(Nm1Segment):
    """
    Claim Rendering Provider Name
    """

    entity_identifier_code: Literal["82"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2310DRefSegment(Nm1Segment):
    """
    Claim Rendering Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310ENm1Segment(Nm1Segment):
    """
    Service Facility Location Name
    """

    entity_identifier_code: Literal["77"]
    entity_type_qualifier: Literal["2"]
    name_last_or_organization_name: Optional[str]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2310ERefSegment(RefSegment):
    """
    Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2310FNm1Segment(Nm1Segment):
    """
    Claim ambulance pickup location
    """

    entity_identifier_code: Literal["DN"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2310FRefSegment(RefSegment):
    """
    Referring Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2320SbrSegment(SbrSegment):
    """
    Claim Other Subscriber
    """

    class PayerResponsibilityCode(str, Enum):
        """
        Code values for SBR01
        """

        PAYER_RESPONSIBILITY_FOUR = "A"
        PAYER_RESPONSIBILITY_FIVE = "B"
        PAYER_RESPONSIBILITY_SIX = "C"
        PAYER_RESPONSIBILITY_SEVEN = "D"
        PAYER_RESPONSIBILITY_EIGHT = "E"
        PAYER_RESPONSIBILITY_NINE = "F"
        PAYER_RESPONSIBILITY_TEN = "G"
        PAYER_RESPONSIBILITY_ELEVEN = "H"
        PAYER_RESPONSIBILITY_PRIMARY = "P"
        PAYER_RESPONSIBILITY_SECONDARY = "S"
        PAYER_RESPONSIBILITY_TERTIARY = "T"
        PAYER_RESPONSIBILITY_UNKNOWN = "U"

    class IndividualRelationshipCode(str, Enum):
        """
        Code values for SBR02
        """

        SPOUSE = "01"
        SELF = "18"
        CHILD = "19"
        EMPLOYEE = "20"
        UNKNOWN = "21"
        ORGAN_DONOR = "39"
        CADAVER_DONOR = "40"
        LIFE_PARTNER = "53"
        OTHER_RELATIONSHIP = "G8"

    class ClaimFilingIndicatorCode(str, Enum):
        """
        SBR09 code values
        """

        OTHER_NON_FEDERAL_PROGRAMS = "11"
        PREFERRED_PROVIDER_ORGANIZATION = "12"
        POINT_OF_SERVICE = "13"
        EXCLUSIVE_PROVIDER_ORGANIZATION = "14"
        INDEMNITY_INSURANCE = "15"
        HEALTH_MAINTENANCE_ORGANIZATION_MEDICARE_RISK = "16"
        DENTAL_MAINTENANCE_ORGANIZATION = "17"
        AUTOMOBILE_MEDICAL = "AM"
        BLUE_CROSS_BLUE_SHIELD = "BL"
        CHAMPUS = "CH"
        COMMERICIAL_INSURANCE_COMPANY = "CI"
        DISABILITY = "DS"
        FEDERAL_EMPLOYEES_PROGRAM = "FI"
        HEALTH_MAINTENANCE_ORGANIZATION = "HM"
        LIABILITY_MEDICAL = "LM"
        MEDICARE_PART_A = "MA"
        MEDICARE_PART_B = "MB"
        MEDICAID = "MC"
        OTHER_FEDERAL_PROGRAM = "OF"
        TITLE_V = "TV"
        VETERANS_AFFAIR_PLAN = "VA"
        WORKERS_COMPENSATION_HEALTH_CLAIM = "WC"
        MUTUALLY_DEFINED = "ZZ"

    payer_responsibility_code: PayerResponsibilityCode
    individual_relationship_code: IndividualRelationshipCode
    claim_filing_indicator_code: ClaimFilingIndicatorCode


class Loop2320AmtSegment(AmtSegment):
    """
    Other Subscriber COB Payer Paid Amount
    """

    class AmountQualifierCode(str, Enum):
        """
        Code values for AMT01
        """

        PAYOR_AMOUNT_PAID = "D"
        AMOUNT_OWED = "EAF"
        NONCOVERED_CHARGES_ACTUAL = "A8"

    amount_qualifier_code: AmountQualifierCode


class Loop2330aNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Name
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        STANDARD_HEALTH_ID_US = "II"
        MEMBER_IDENTIFICATION_NUMBER = "MI"

    entity_identifier_code: Literal["IL"]
    identification_code_qualifier: IdentificationCodeQualifier
    identification_code: str = Field(min_length=2, max_length=80)


class Loop2330aRefSegment(RefSegment):
    """
    Claim Other Subscriber Reference Identification
    """

    reference_identification_qualifier: Literal["SY"]


class Loop2330bNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Name
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        PAYOR_IDENTIFICATION = "PI"
        CMS_PLAN_ID = "XV"

    entity_identifier_code: Literal["PR"]
    entity_type_qualifier: Literal["2"]
    name_last_or_organization_name: str = Field(min_length=1, max_length=60)
    identification_code_qualifier: IdentificationCodeQualifier
    identification_code: str = Field(min_length=2, max_length=80)


class Loop2330BDtpSegment(DtpSegment):
    """
    Claim Other Subscriber Other Payer Claim Check or Remittance Date
    """

    date_time_qualifier: Literal["573"]
    date_time_period_format_qualifier: Literal["D8"]


class Loop2300BRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYER_IDENTIFICATION_NUMBER = "2U"
        EMPLOYER_IDENTIFICATION_NUMBER = "EI"
        CLAIM_OFFICE_NUMBER = "FY"
        NAIC_CODE = "NF"
        PRIOR_AUTHORIZATION_NUMBER = "G1"
        REFERRAL_NUMBER = "9F"
        SIGNAL_CODE = "T4"
        ORIGINAL_REFERENCE_NUMBER = "F8"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330cNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Attending Provider Name
    """

    entity_identifier_code: Literal["71"]
    entity_type_qualifier: Literal["1"]
    name_last_or_organization_name: Optional[str]


class Loop2330cRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Attending Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330dNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Operating Physician
    """

    entity_identifier_code: Literal["72"]
    entity_type_qualifier: Literal["1"]
    name_last_or_organization_name: Optional[str]


class Loop2330dRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Operating Physician Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330eNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Other Operating Physician
    """

    entity_identifier_code: Literal["ZZ"]
    entity_type_qualifier: Literal["1"]
    name_last_or_organization_name: Optional[str]


class Loop2330eRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Other Operating Physician Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330fNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Service Facility Location
    """

    entity_identifier_code: Literal["77"]
    entity_type_qualifier: Literal["2"]
    name_last_or_organization_name: Optional[str]


class Loop2330fRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Service Facility Location Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330gNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Rendering Provider Name
    """

    entity_identifier_code: Literal["82"]
    entity_type_qualifier: Literal["1"]
    name_last_or_organization_name: Optional[str]


class Loop2330gRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Rendering Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330HNm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Referring Provider Name
    """

    entity_identifier_code: Literal["DN"]
    entity_type_qualifier: Literal["1"]
    name_last_or_organization_name: Optional[str]


class Loop2330HRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Referring Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2330INm1Segment(Nm1Segment):
    """
    Claim Other Subscriber Other Payer Billing Provider Name
    """

    entity_identifier_code: Literal["85"]
    entity_type_qualifier: Literal["2"]
    name_last_or_organization_name: Optional[str]


class Loop2330IRefSegment(RefSegment):
    """
    Claim Other Subscriber Other Payer Billing Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2400DtpSegment(DtpSegment):
    """
    Date values for Loop 2400 - Claim service line
    """

    date_time_qualifier: Literal["472"]
    date_time_period_format_qualifier: Literal["D8"]


class Loop2400PwkSegment(PwkSegment):
    """
    Service line paperwork segment
    """

    class AttachmentReportTypeCode(str, Enum):
        """
        PWK01 code values
        """

        REPORT_JUSTIFYING_TREATMENT_BEYOND_UTILIZATION_GUIDELINES = "03"
        DRUGS_ADMINISTERED = "04"
        TREATMENT_DIAGNOSIS = "05"
        INITIAL_ASSESSMENT = "06"
        FUNCTIONAL_GOALS = "07"
        PLAN_OF_TREATMENT = "08"
        PROGRESS_REPORT = "09"
        CONTINUED_TREATMENT = "10"
        CHEMICAL_ANALYSIS = "11"
        CERTIFIED_TEST_REPORT = "13"
        JUSTIFICATION_FOR_ADMISSION = "15"
        RECOVERY_PLAN = "21"
        ALLERGIES_SENSITIVITIES_DOCUMENT = "A3"
        AUTOPSY_REPORT = "A4"
        AMBULANCE_CERTIFICATION = "AM"
        ADMISSION_SUMMARY = "AS"
        PRESCRIPTION = "B2"
        PHYSICIAN_ORDER = "B3"
        REFERRAL_FORM = "B4"
        BENCHMARK_TESTING_RESULTS = "BR"
        BASELINE = "BS"
        BLANKET_TESTING_RESULTS = "BT"
        CHIROPRACTIC_JUSTIFICATION = "CB"
        CONSENT_FORM = "CK"
        CERTIFICATION = "CT"
        DRUG_PROFILE_DOCUMENT = "D2"
        DENTAL_MODELS = "DA"
        DURABLE_MEDICAL_EQUIPMENT_PRESCRIPTION = "DB"
        DIAGNOSTIC_REPORT = "DG"
        DISCHARGE_MONITORING_REPORT = "DJ"
        DISCHARGE_SUMMARY = "DS"
        EXPLANATION_OF_BENEFITS = "EB"
        HEALTH_CERTIFICATE = "HC"
        HEALTH_CLINIC_RECORDS = "HR"
        IMMUNIZATION_RECORD = "I5"
        STATE_SCHOOL_IMMUNIZATION_RECORDS = "IR"
        LABORATORY_RESULTS = "LA"
        MEDICAL_RECORD_ATTACHMENT = "M1"
        MODELS = "MT"
        NURSING_NOTES = "NN"
        OPERATIVE_NOTE = "OB"
        OXYGEN_CONTENT_AVERAGING_REPORT = "OC"
        ORDERS_AND_TREATMENTS_DOCUMENT = "OD"
        OBJECTIVE_PHYSICAL_EXAMINATION = "OE"
        OXYGEN_THERAPY_CERTIFICATION = "OX"
        SUPPORT_DATA_FOR_CLAIM = "OZ"
        PATHOLOGY_REPORT = "P4"
        PATIENT_MEDICAL_HISTORY_DOCUMENT = "P5"
        PARENTAL_OR_ENTERAL_CERTIFICATION = "PE"
        PHYSICAL_THERAPY_NOTES = "PN"
        PROSTHETICS_OR_ORTHOTIC_CERTIFICATION = "PO"
        PARAMEDICAL_RESULTS = "PQ"
        PHYSICIANS_REPORT = "PY"
        PHYSICAL_THERAPY_CERTIFICATION = "PZ"
        RADIOLOGY_FILMS = "RB"
        RADIOLOGY_REPORTS = "RR"
        REPORT_OF_TESTS_AND_ANALYSIS_REPORT = "RT"
        RENEWABLE_OXYGEN_CONTENT_AVERAGING_REPORT = "RX"
        SYMPTOMS_DOCUMENT = "SG"
        DEATH_NOTIFICATION = "V5"
        PHOTOGRAPHS = "XP"

    class AttachmentTransmissionCode(str, Enum):
        """
        Code values for PWK02
        """

        AVAILABLE_ON_REQUEST_PROVIDER_SITE = "AA"
        BY_MAIL = "BM"
        ELECTRONICALLY_ONLY = "EL"
        EMAIL = "EM"
        FILE_TRANSFER = "FT"
        BY_FAX = "FX"


class Loop2400RefSegment(RefSegment):
    """
    Loop 2400/Service Line Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PROVIDER_CONTROL_NUMBER = "6R"
        REPRICED_LINE_ITEM_REFERENCE_NUMBER = "9B"
        ADJUSTED_REPRICED_LINE_ITEM_REFERENCE_NUMBER = "9D"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2400AmtSegment(AmtSegment):
    """
    Loop 2400/Service Line Amounts
    """

    class AmountQualifierCode(str, Enum):
        """
        Code values for AMT01
        """

        GOODS_SERVICES_TAX = "GT"
        MISCELLANEOUS_TAXES = "N8"

    amount_qualifier_code: AmountQualifierCode


class Loop2400NteSegment(NteSegment):
    """
    Loop 2400/Claim Service Line Notes
    """

    note_reference_code: Literal["TPO"]


class Loop2410RefSegment(RefSegment):
    """
    Loop 2400/Claim Service Line Drug Identification Reference Identifier
    """

    reference_identification_qualifier: Literal["XZ"]


class Loop2420ANm1Segment(Nm1Segment):
    """
    Loop 2420A Operating Physician Name
    """

    entity_identifier_code: Literal["71"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Literal["XX"]


class Loop2420ARefSegment(RefSegment):
    """
    Loop 2420A Operating Physician Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420BNm1Segment(Nm1Segment):
    """
    Loop 2420B Other Operating Physician Name
    """

    entity_identifier_code: Literal["ZZ"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2420BRefSegment(RefSegment):
    """
    Loop 2420B Purchased Service Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420CNm1Segment(Nm1Segment):
    """
    Loop 2420C Rendering Provider Name
    """

    entity_identifier_code: Literal["82"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2420CRefSegment(RefSegment):
    """
    Loop 2420C Rendering Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2420DNm1Segment(Nm1Segment):
    """
    Loop 2420D Referring Provider Name
    """

    entity_identifier_code: Literal["DN"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[Literal["XX"]]


class Loop2420DRefSegment(RefSegment):
    """
    Loop 2420D Referring Provider Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        PROVIDER_UPIN_NUMBER = "1G"
        PROVIDER_COMMERCIAL_NUMBER = "G2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2430DtpSegment(DtpSegment):
    """
    Loop 2430 Line Check or Remittance Date
    """

    date_time_qualifier: Literal["573"]
    date_time_period_format_qualifier: Literal["D8"]


class Loop2430AmtSegment(AmtSegment):
    """
    Loop 2430 Remaining Patient Liability
    """

    amount_qualifier_code: Literal["EAF"]



"""
test_837_005010XX223A3.py
"""
import pytest
from tests.support import assert_eq_model, resources_directory
import os


@pytest.fixture
def resource_path() -> str:
    return os.path.join(resources_directory, "837_005010X223A3")


@pytest.mark.parametrize(
    "file_name",
    [
        "institutional-claim.837i",
        "out-of-network-repriced-claim.837i",
        "ppo-repriced-claim.837i",
        "two-claims-single-provider.837i",
    ],
)
def test_837i_model(resource_path, file_name: str):
    x12_file_path = os.path.join(resource_path, file_name)
    assert os.path.exists(x12_file_path)
    assert_eq_model(x12_file_path)



"""
loops.py

Models the loops, or logical segment groupings, for the Eligibility 271 005010X279A1 transaction set.
The Eligibility Transaction set organizes loops into a hierarchical and nested model.

- Header
- Loop 2000A (Information Source)
    -- Loop 2100A (Information Source Name)
    -- Loop 2000B (Information Receiver)
        --- Loop 2100B (Information Receiver Name)
        --- Loop 2000C (Subscriber)
            --- Loop 2100C (Subscriber Name)
                --- Loop 2110C (Subscriber Eligibility)
            --- Loop 2000D (Dependent)
                --- Loop 2100D (Dependent Name)
                    --- Loop 2110D (Dependent Eligibility)
-- Footer

The Header and Footer components are not "loops" per the specification, but are included to standardize and simplify
transactional modeling and processing.
"""

from x12.models import X12SegmentGroup
from x12.segments import (
    SeSegment,
    N3Segment,
    N4Segment,
    TrnSegment,
    DmgSegment,
    HiSegment,
    MpiSegment,
    EbSegment,
    HsdSegment,
    MsgSegment,
    IiiSegment,
    LsSegment,
    LeSegment,
    PerSegment,
    PrvSegment,
)
from .segments import (
    HeaderStSegment,
    HeaderBhtSegment,
    Loop2000CHlSegment,
    Loop2100BNm1Segment,
    Loop2000BHlSegment,
    Loop2100BRefSegment,
    Loop2100BPrvSegment,
    Loop2100ANm1Segment,
    Loop2000AHlSegment,
    Loop2000AAaaSegment,
    Loop2100CNm1Segment,
    Loop2100DNm1Segment,
    Loop2100RefSegment,
    Loop2100CInsSegment,
    Loop2100DInsSegment,
    Loop2100DtpSegment,
    Loop2110RefSegment,
    Loop2110DtpSegment,
    Loop2100AAaaSegment,
    Loop2100BAaaSegment,
    Loop2110CAaaSegment,
    Loop2120Nm1Segment,
    Loop2000DHlSegment,
)
from typing import List, Optional
from pydantic import Field, root_validator
from x12.validators import validate_duplicate_ref_codes


class Header(X12SegmentGroup):
    """
    Transaction Header Information
    """

    st_segment: HeaderStSegment
    bht_segment: HeaderBhtSegment


class Loop2120D(X12SegmentGroup):
    """
    Loop 2120D
    """

    nm1_segment: Loop2120Nm1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    per_segment: Optional[List[PerSegment]] = Field(min_items=0, max_items=3)
    prv_segment: Optional[PrvSegment]


class Loop2115D(X12SegmentGroup):
    """
    Loop 2115D - Dependent Eligibility or Benefit Additional Information
    """

    iii_segment: IiiSegment
    ls_segment: LsSegment
    loop_2120c: Optional[List[Loop2120D]] = Field(min_items=0, max_items=23)
    le_segment: LeSegment


class Loop2110D(X12SegmentGroup):
    """
    Loop 2110D Dependent Eligibility or Benefit Information
    """

    eb_segment: EbSegment
    hsd_segment: Optional[List[HsdSegment]] = Field(min_items=0, max_items=9)
    ref_segment: Optional[List[Loop2110RefSegment]] = Field(min_items=0, max_items=9)
    dtp_segment: Optional[List[Loop2110DtpSegment]] = Field(min_items=0, max_items=20)
    aaa_segment: Optional[List[Loop2110CAaaSegment]] = Field(min_items=0, max_items=9)
    msg_segment: Optional[List[MsgSegment]] = Field(min_items=0, max_items=10)
    loop_2115d: Optional[List[Loop2115D]] = Field(min_items=0, max_items=10)
    ls_segment: Optional[LsSegment]
    loop_2120d: Optional[List[Loop2120D]] = Field(min_items=0, max_items=23)
    le_segment: Optional[LeSegment]

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )


class Loop2100D(X12SegmentGroup):
    """
    Loop 2100D - Dependent Name
    """

    nm1_segment: Loop2100DNm1Segment
    ref_segment: Optional[List[Loop2100RefSegment]] = Field(min_items=0, max_items=9)
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    # Loop2100D AAA is identical to Loop2100B AAA
    aaa_segment: Optional[List[Loop2100BAaaSegment]] = Field(min_items=0, max_items=9)
    prv_segment: Optional[PrvSegment]
    dmg_segment: Optional[DmgSegment]
    ins_segment: Optional[Loop2100DInsSegment]
    hi_segment: Optional[HiSegment]
    dtp_segment: Optional[List[Loop2100DtpSegment]] = Field(min_items=0, max_items=9)
    mpi_segment: Optional[MpiSegment]
    loop_2110d: Optional[List[Loop2110D]] = Field(min_items=0)

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )


class Loop2000D(X12SegmentGroup):
    """
    Loop 2000D - Dependent
    """

    hl_segment: Loop2000DHlSegment
    trn_segment: Optional[List[TrnSegment]] = Field(min_items=0, max_items=2)
    loop_2100d: Loop2100D


class Loop2120C(X12SegmentGroup):
    """
    Loop 2120C Subscriber Benefit Related Entity Name
    """

    nm1_segment: Loop2120Nm1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    per_segment: Optional[List[PerSegment]] = Field(min_items=0, max_items=3)
    prv_segment: Optional[PrvSegment]


class Loop2115C(X12SegmentGroup):
    """
    Loop 2115C - Subscriber Eligibility or Benefit Information Additional Information
    """

    iii_segment: IiiSegment
    ls_segment: LsSegment
    loop_2120c: Optional[List[Loop2120C]] = Field(min_items=0, max_items=23)
    le_segment: LeSegment


class Loop2110C(X12SegmentGroup):
    """
    Loop2110C - Subscriber Eligibility or Benefit Information
    """

    eb_segment: EbSegment
    hsd_segment: Optional[List[HsdSegment]] = Field(min_items=0, max_items=9)
    ref_segment: Optional[List[Loop2110RefSegment]] = Field(min_items=0, max_items=9)
    dtp_segment: Optional[List[Loop2110DtpSegment]] = Field(min_items=0, max_items=20)
    aaa_segment: Optional[List[Loop2110CAaaSegment]] = Field(min_items=0, max_items=9)
    msg_segment: Optional[List[MsgSegment]] = Field(min_items=0, max_items=10)
    loop_2115c: Optional[List[Loop2115C]] = Field(min_items=0, max_items=10)
    ls_segment: Optional[LsSegment]
    loop_2120c: Optional[List[Loop2120C]] = Field(min_items=0, max_items=23)
    le_segment: Optional[LeSegment]

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )

    @root_validator
    def validate_red_cross_eb_ref_codes(cls, values):
        """
        Validates that reference identification codes are limited when American Red Cross is the eligibility benefit.

        :@param values: The validated model values
        """
        benefit_code = values["eb_segment"].eligibility_benefit_information
        arc_ref_types = {"1W", "49", "F6", "NQ"}
        ref_types = {
            r.reference_identification_qualifier for r in values.get("ref_segments", [])
        }

        if ref_types and benefit_code == "R" and (ref_types - arc_ref_types):
            raise ValueError(f"{ref_types} are not valid for American Red Cross")

        return values


class Loop2100C(X12SegmentGroup):
    """
    Loop 2100C - Subscriber Name
    """

    nm1_segment: Loop2100CNm1Segment
    ref_segment: Optional[List[Loop2100RefSegment]] = Field(min_items=0, max_items=9)
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    # Loop2100C AAA is identical to Loop2100B AAA
    aaa_segment: Optional[List[Loop2100BAaaSegment]] = Field(min_items=0, max_items=9)
    prv_segment: Optional[PrvSegment]
    dmg_segment: Optional[DmgSegment]
    ins_segment: Optional[Loop2100CInsSegment]
    hi_segment: Optional[HiSegment]
    dtp_segment: Optional[List[Loop2100DtpSegment]] = Field(min_items=0, max_items=9)
    mpi_segment: Optional[MpiSegment]
    loop_2110c: Optional[List[Loop2110C]] = Field(min_items=0)

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )


class Loop2000C(X12SegmentGroup):
    """
    Loop 2000C - Subscriber
    """

    hl_segment: Loop2000CHlSegment
    trn_segment: Optional[List[TrnSegment]] = Field(min_items=0, max_items=2)
    loop_2100c: Loop2100C
    loop_2000d: Optional[List[Loop2000D]] = Field(min_items=0)


class Loop2100B(X12SegmentGroup):
    """
    Loop 2100B - Information Receiver Name
    """

    nm1_segment: Loop2100BNm1Segment
    ref_segment: Optional[List[Loop2100BRefSegment]]
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    aaa_segment: Optional[List[Loop2100BAaaSegment]]
    prv_segment: Optional[Loop2100BPrvSegment]

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )


class Loop2000B(X12SegmentGroup):
    """
    Loop 2000B - Information Receiver
    """

    hl_segment: Loop2000BHlSegment
    loop_2100b: Loop2100B
    loop_2000c: Optional[List[Loop2000C]]


class Loop2100A(X12SegmentGroup):
    """
    Loop 2100A - Information Source Name
    """

    nm1_segment: Loop2100ANm1Segment
    prv_segment: Optional[List[PrvSegment]] = Field(min_items=0, max_items=3)
    aaa_segment: Optional[List[Loop2100AAaaSegment]] = Field(min_items=0, max_items=9)


class Loop2000A(X12SegmentGroup):
    """
    Loop 2000A - Information Source
    The root node/loop for the 271 transaction

    """

    hl_segment: Loop2000AHlSegment
    aaa_segment: Optional[List[Loop2000AAaaSegment]] = Field(min_items=0, max_items=9)
    loop_2100a: Loop2100A
    loop_2000b: List[Loop2000B] = Field(min_items=0)


class Footer(X12SegmentGroup):
    """
    Transaction Footer Information
    """

    se_segment: SeSegment



"""
support.py

Reusable assertions and utilities for X12 based test cases
"""
from x12.io import X12ModelReader
import os

# base resource directory for "file fixtures" (sample x12 transactions)
resources_directory = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "resources",
)


def assert_eq_model(x12_path: str):
    """
    Asserts that a generated X12 model is equal to provided input.
    `assert_eq_model` is constrained to processing an input with a single functional group, but does support
    multiple transaction sets.

    :param x12_path: The path to a X12 file
    """
    with open(x12_path) as f:
        x12_data = "".join([l for l in f.readlines() if l])

        with X12ModelReader(x12_data) as r:

            segment_terminator = r._x12_segment_reader.delimiters.segment_terminator
            segments = [
                s for s in x12_data.split(segment_terminator) if s and s != "\n"
            ]

            assert "ISA" in segments[0]
            assert "GS" in segments[1]
            assert "GE" in segments[-2]
            assert "IEA" in segments[-1]

            control_header = (
                segment_terminator.join(segments[0:2]) + segment_terminator + "\n"
            )
            control_footer = segment_terminator.join(segments[-2:]) + segment_terminator
            transaction_sets: str = ""

            for m in r.models():
                transaction_sets += m.x12()

            actual = f"{control_header}{transaction_sets}{control_footer}"
            assert actual == x12_data



"""
transaction_set.py

Defines the Health Care Claims Status 276 005010X212 transaction set model.
"""
from x12.models import X12SegmentGroup
from .loops import Header, Footer, Loop2000A
from typing import List
from pydantic import Field, root_validator
from x12.validators import validate_segment_count


class HealthCareClaimsStatusRequest(X12SegmentGroup):
    """
    The Health Care Claims Status transaction model - 276
    """

    header: Header
    loop_2000a: List[Loop2000A] = Field(min_items=1)
    footer: Footer

    _validate_segment_count = root_validator(allow_reuse=True)(validate_segment_count)



"""
transactions.py

The transactions module serves as the package root for X12 transaction implementations.
Transaction implementations are stored in separate packages with the following structure:

    x12_[transaction_code]_[implementation_version]
    / loops.py - Transaction Loop Domain Models
    / parsing.py - Transaction Set Parser Implementation and Loop Parser Functions
    / segments.py - Specialized Segments for Loop Domain Models
    / transaction_set.py - Transaction Set Domain Model
"""



"""
loops.py

Models the loops, or logical segment groupings, for the Health Care Claim Payment 835 005010X221A1 transaction set model.
The Health Care Claim Payment transaction set organizes loops into a hierarchical and nested model.
"""
from x12.models import X12SegmentGroup
from x12.segments import (
    SeSegment,
    BprSegment,
    DtmSegment,
    N3Segment,
    N4Segment,
    RdmSegment,
    LxSegment,
    Ts3Segment,
    Ts2Segment,
    ClpSegment,
    Nm1Segment,
    MiaSegment,
    MoaSegment,
    SvcSegment,
    PlbSegment,
)
from .segments import (
    HeaderStSegment,
    HeaderTrnSegment,
    HeaderCurSegment,
    HeaderRefSegment,
    Loop1000AN1Segment,
    Loop1000ARefSegment,
    Loop1000APerSegment,
    Loop1000BN1Segment,
    Loop1000BRefSegment,
    Loop2100CasSegment,
    Loop2100RefSegment,
    Loop2100DtmSegment,
    Loop2100PerSegment,
    Loop2100AmtSegment,
    Loop2100QtySegment,
    Loop2110DtmSegment,
    Loop2110CasSegment,
    Loop2110RefSegment,
    Loop2110AmtSegment,
    Loop2110QtySegment,
    Loop2110LqSegment,
)
from typing import Optional, List
from pydantic import Field, root_validator
from decimal import Decimal


class Header(X12SegmentGroup):
    """
    Transaction Header Information
    """

    st_segment: HeaderStSegment
    bpr_segment: BprSegment
    trn_segment: HeaderTrnSegment
    cur_segment: Optional[HeaderCurSegment]
    ref_segment: Optional[List[HeaderRefSegment]] = Field(min_items=0, max_items=2)
    dtm_segment: Optional[DtmSegment]


class Loop1000A(X12SegmentGroup):
    """
    Loop 1000A - Payer Identification
    """

    n1_segment: Loop1000AN1Segment
    n3_segment: N3Segment
    n4_segment: N4Segment
    ref_segment: Optional[List[Loop1000ARefSegment]] = Field(min_items=0, max_items=4)
    per_segment: Optional[List[Loop1000APerSegment]]


class Loop1000B(X12SegmentGroup):
    """
    Loop 1000B - Payee Identification
    """

    n1_segment: Loop1000BN1Segment
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    ref_segment: Optional[List[Loop1000BRefSegment]]
    rdm_segment: Optional[RdmSegment]


class Loop2110(X12SegmentGroup):
    """
    Loop 2110 - Service Line Payment Information
    """

    svc_segment: SvcSegment
    dtm_segment: Optional[List[Loop2110DtmSegment]] = Field(min_items=0, max_items=2)
    cas_segment: Optional[List[Loop2110CasSegment]] = Field(min_items=0, max_items=99)
    ref_segment: Optional[List[Loop2110RefSegment]] = Field(min_items=0, max_items=24)
    amt_segment: Optional[List[Loop2110AmtSegment]] = Field(min_items=0, max_items=9)
    qty_segment: Optional[List[Loop2110QtySegment]] = Field(min_items=0, max_items=6)
    lq_segment: Optional[List[Loop2110LqSegment]] = Field(min_items=0, max_items=99)


class Loop2100(X12SegmentGroup):
    """
    Loop 2100 - Claim Payment Information
    """

    clp_segment: ClpSegment
    cas_segment: Optional[List[Loop2100CasSegment]] = Field(min_items=0, max_items=99)
    nm1_segment: List[Nm1Segment] = Field(min_items=1, max_items=7)
    mia_segment: Optional[MiaSegment]
    moa_segment: Optional[MoaSegment]
    ref_segment: Optional[List[Loop2100RefSegment]] = Field(min_items=0, max_items=15)
    dtm_segment: Optional[List[Loop2100DtmSegment]] = Field(min_items=0, max_items=5)
    per_segment: Optional[List[Loop2100PerSegment]] = Field(min_items=0, max_items=2)
    amt_segment: Optional[List[Loop2100AmtSegment]] = Field(min_items=0, max_items=13)
    qty_segment: Optional[List[Loop2100QtySegment]] = Field(min_items=0, max_items=14)
    loop_2110: Optional[List[Loop2110]] = Field(min_items=0, max_items=99)

    @root_validator
    def validate_balance(cls, values):
        """
        Validates the claim totals reported in the CLP segment against the adjustments made in CAS segments.
        The balance is calculated as:
            clp.charge amount - cas.adjustments = clp.payment_amount

        CAS segments exist within loops 2100 and 2110
        """
        clp_segment = values.get("clp_segment")
        if not clp_segment:
            return values
        charge_amount = clp_segment.total_claim_charge_amount
        payment_amount = clp_segment.claim_payment_amount
        adjustment_amount = Decimal("0.0")

        cas_segments = values.get("cas_segment", [])
        for adjustment in cas_segments:
            adjustment_data = adjustment.dict()
            for i in range(1, 7, 1):
                amount = adjustment_data.get(f"monetary_amount_{i}")
                adjustment_amount += amount if amount else Decimal("0.0")

        loop_2110 = values.get("loop_2110") or []
        for service_payment in loop_2110:
            adjustments = service_payment.cas_segment or []
            for adjustment in adjustments:
                adjustment_data = adjustment.dict()
                for i in range(1, 7, 1):
                    amount = adjustment_data.get(f"monetary_amount_{i}")
                    adjustment_amount += amount if amount else Decimal("0.0")

        if charge_amount - payment_amount != adjustment_amount:
            raise ValueError(
                f"Unable to balance charge amount {charge_amount} paid amount {payment_amount} against adjustments {adjustment_amount}"
            )

        return values


class Loop2000(X12SegmentGroup):
    """
    Loop 2000 - Claim Payment Header Number
    """

    lx_segment: LxSegment
    ts3_segment: Optional[Ts3Segment]
    ts2_segment: Optional[Ts2Segment]
    loop_2100: List[Loop2100]


class Footer(X12SegmentGroup):
    """
    Transaction Footer Information
    """

    plb_segment: Optional[PlbSegment]
    se_segment: SeSegment



"""
test_835_005010XX221A1.py
"""
import pytest
from tests.support import assert_eq_model, resources_directory
import os


@pytest.fixture
def resource_path() -> str:
    return os.path.join(resources_directory, "835_005010X221A1")


@pytest.mark.parametrize(
    "file_name",
    [
        "medicare-part-a.835",
        "managed-care.835",
        "secondary-payment.835",
        "tertiary-payment.835",
        "cob-contractural-adjustment.835",
    ],
)
def test_835_model(resource_path, file_name: str):
    x12_file_path = os.path.join(resource_path, file_name)
    assert os.path.exists(x12_file_path)
    assert_eq_model(x12_file_path)



"""
loops.py

Models the loops, or logical segment groupings, for the Health Care Claim Status Response 277 005010X212 transaction set.
The Health Care Claim Status Response organizes loops into a hierarchical and nested model.

-- Header
    -- Loop 2000A (Information Source)
        -- Loop 2100A (Payer Name)
        -- Loop 2000B (Information Receiver)
            -- Loop 2100B (Information Receiver Name)
            -- Loop 2200B (Information Receiver Trace Identifier)
            -- Loop 2000C (Service Provider Detail)
                -- Loop 2100C (Service Provider Name)
                -- Loop 2200C (Information Receiver Trace Identifier)
                -- Loop 2000D (Subscriber Level)
                    -- Loop 2100D (Subscriber Name)
                    -- Loop 2200D (Claim Status Tracking Number)
                        -- Loop 2220D (Service Line Information)
                    -- Loop 2000E (Dependent Level)
                        -- Loop 2100E (Dependent Name)
                        -- Loop 2200E (Claim Status Tracking Number)
                            -- Loop 2220E (Service Line Information)
-- Footer

The Subscriber and Dependent Claim Status and Service Line Loops share segment classes as the data structures
are identical.
"""
from x12.models import X12SegmentGroup
from .segments import (
    HeaderStSegment,
    HeaderBhtSegment,
    Loop2000AHlSegment,
    Loop2100ANm1Segment,
    Loop2100APerSegment,
    Loop2000BHlSegment,
    Loop2100BNm1Segment,
    Loop2200BTrnSegment,
    Loop2200CTrnSegment,
    Loop2000CHlSegment,
    Loop2100CNm1Segment,
    Loop2000DHlSegment,
    Loop2100DNm1Segment,
    Loop2200DTrnSegment,
    Loop2200DRefSegment,
    Loop2200DDtpSegment,
    Loop2220DSvcSegment,
    Loop2220DRefSegment,
    Loop2220DDtpSegment,
    Loop2000EHlSegment,
    Loop2100ENm1Segment,
)
from x12.segments import SeSegment, StcSegment
from typing import List, Optional
from pydantic import Field


class Header(X12SegmentGroup):
    """
    Transaction Header Information
    """

    st_segment: HeaderStSegment
    bht_segment: HeaderBhtSegment


class Loop2220E(X12SegmentGroup):
    """
    Dependent Claim Status Service Line Information
    The segment implementations from the Subscriber loop are used as the structure is the same.
    """

    svc_segment: Loop2220DSvcSegment
    stc_segment: List[StcSegment]
    ref_segment: Optional[Loop2220DRefSegment]
    dtp_segment: Loop2220DDtpSegment


class Loop2200E(X12SegmentGroup):
    """
    Dependent Claim Status Tracking Number
    The segment implementations from the Subscriber loop are used as the structure is the same.
    """

    trn_segment: Loop2200DTrnSegment
    stc_segment: List[StcSegment]
    ref_segment: Optional[List[Loop2200DRefSegment]] = Field(min_items=0, max_items=7)
    dtp_segment: Optional[Loop2200DDtpSegment]
    loop_2220e: Optional[List[Loop2220E]]


class Loop2100E(X12SegmentGroup):
    """
    Dependent Name
    """

    nm1_segment: Loop2100ENm1Segment


class Loop2000E(X12SegmentGroup):
    """
    Dependent Loop
    """

    hl_segment: Loop2000EHlSegment
    loop_2100e: Loop2100E
    loop_2200e: Optional[List[Loop2200E]]


class Loop2220D(X12SegmentGroup):
    """
    Subscriber Claim Status Service Line Information
    """

    svc_segment: Loop2220DSvcSegment
    stc_segment: List[StcSegment]
    ref_segment: Optional[Loop2220DRefSegment]
    dtp_segment: Loop2220DDtpSegment


class Loop2200D(X12SegmentGroup):
    """
    Subscriber Claim Status Tracking Number
    """

    trn_segment: Loop2200DTrnSegment
    stc_segment: List[StcSegment]
    ref_segment: Optional[List[Loop2200DRefSegment]] = Field(min_items=0, max_items=7)
    dtp_segment: Optional[Loop2200DDtpSegment]
    loop_2220d: Optional[List[Loop2220D]]


class Loop2100D(X12SegmentGroup):
    """
    Loop 2100D Subscriber Name
    """

    nm1_segment: Loop2100DNm1Segment


class Loop2000D(X12SegmentGroup):
    """
    Loop 2000D Subscriber
    """

    hl_segment: Loop2000DHlSegment
    loop_2100d: Loop2100D
    loop_2200d: Optional[List[Loop2200D]]
    loop_2000e: Optional[List[Loop2000E]]


class Loop2200C(X12SegmentGroup):
    """
    Loop 2200C - Service Provider Trace Trace Identifiers
    The Loop2200BTrnSegment is used as it's structure is compatible with Loop2200C
    """

    trn_segment: Loop2200CTrnSegment
    stc_segment: Optional[List[StcSegment]]


class Loop2100C(X12SegmentGroup):
    """
    Loop 2100C - Service Provider Name
    """

    nm1_segment: Loop2100CNm1Segment


class Loop2000C(X12SegmentGroup):
    """
    Loop 2000C - Service Provider
    """

    hl_segment: Loop2000CHlSegment
    loop_2100c: Loop2100C
    loop_2200c: Optional[Loop2200C]
    loop_2000d: Optional[List[Loop2000D]] = Field(min_items=1)


class Loop2200B(X12SegmentGroup):
    """
    Loop 2200B - Information Receiver Trace Identifiers
    """

    trn_segment: Loop2200BTrnSegment
    stc_segment: Optional[List[StcSegment]]


class Loop2100B(X12SegmentGroup):
    """
    Loop 2100B - Information Receiver Name
    """

    nm1_segment: Loop2100BNm1Segment


class Loop2000B(X12SegmentGroup):
    """
    Loop 2000B - Information Receiver
    """

    hl_segment: Loop2000BHlSegment
    loop_2100b: Loop2100B
    loop_2200b: Optional[Loop2200B]
    loop_2000c: Optional[List[Loop2000C]] = Field(min_items=1)


class Loop2100A(X12SegmentGroup):
    """
    Loop 2100A - Payer Name
    """

    nm1_segment: Loop2100ANm1Segment
    per_segment: Optional[Loop2100APerSegment]


class Loop2000A(X12SegmentGroup):
    """
    Loop 2000A - Information Source
    """

    hl_segment: Loop2000AHlSegment
    loop_2100a: Loop2100A
    loop_2000b: List[Loop2000B] = Field(min_items=1)


class Footer(X12SegmentGroup):
    """
    Transaction Footer Information
    """

    se_segment: SeSegment



"""
parsing.py

Parses X12 835 005010X221A1 segments into a transactional domain model.

The parsing module includes a specific parser for the 835 transaction and loop parsing functions to create new loops
as segments are streamed to the transactional data model.

Loop parsing functions are implemented as set_[description]_loop(context: X12ParserContext, segment_data: Dict).
"""
from enum import Enum
from x12.parsing import X12ParserContext, match
from typing import Dict


class TransactionLoops(str, Enum):
    """
    The loops used to support the 835 005010X221A1 format.
    """

    HEADER = "header"
    PAYER_IDENTIFICATION = "loop_1000a"
    PAYEE_IDENTIFICATION = "loop_1000b"
    CLAIM_PAYMENT_LINE_ITEM = "loop_2000"
    CLAIM_PAYMENT_INFORMATION = "loop_2100"
    SERVICE_PAYMENT_INFORMATION = "loop_2110"
    FOOTER = "footer"


def _get_header(context: X12ParserContext) -> Dict:
    """Returns the 837 transaction header"""
    return context.transaction_data[TransactionLoops.HEADER]


def _get_claim_payment_line(context: X12ParserContext) -> Dict:
    """Returns the current claim payment line record"""
    return context.transaction_data[TransactionLoops.CLAIM_PAYMENT_LINE_ITEM][-1]


def _get_claim_payment_information(context: X12ParserContext) -> Dict:
    """Returns the current claim payment info"""
    claim_payment_line = _get_claim_payment_line(context)
    return claim_payment_line[TransactionLoops.CLAIM_PAYMENT_INFORMATION][-1]


@match("ST")
def set_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set header loop for the 835 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    context.transaction_data[TransactionLoops.HEADER] = {"ref_segment": []}
    header = context.transaction_data[TransactionLoops.HEADER]
    context.set_loop_context(TransactionLoops.HEADER, header)


@match("N1", conditions={"entity_identifier_code": "PR"})
def set_payer_identification_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the payer identification loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    context.transaction_data[TransactionLoops.PAYER_IDENTIFICATION] = {
        "ref_segment": []
    }
    loop_data = context.transaction_data[TransactionLoops.PAYER_IDENTIFICATION]
    context.set_loop_context(TransactionLoops.PAYER_IDENTIFICATION, loop_data)


@match("N1", conditions={"entity_identifier_code": "PE"})
def set_payee_identification_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the payee identification loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    context.transaction_data[TransactionLoops.PAYEE_IDENTIFICATION] = {
        "ref_segment": []
    }
    loop_data = context.transaction_data[TransactionLoops.PAYEE_IDENTIFICATION]
    context.set_loop_context(TransactionLoops.PAYEE_IDENTIFICATION, loop_data)


@match("LX")
def set_claim_payment_line_item_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the header number loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if TransactionLoops.CLAIM_PAYMENT_LINE_ITEM not in context.transaction_data:
        context.transaction_data[TransactionLoops.CLAIM_PAYMENT_LINE_ITEM] = [{}]
    else:
        context.transaction_data[TransactionLoops.CLAIM_PAYMENT_LINE_ITEM].append({})

    loop_data = context.transaction_data[TransactionLoops.CLAIM_PAYMENT_LINE_ITEM][-1]
    context.set_loop_context(TransactionLoops.CLAIM_PAYMENT_LINE_ITEM, loop_data)


@match("CLP")
def set_claim_payment_information_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the claim payment loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    claim_payment_line = _get_claim_payment_line(context)

    if TransactionLoops.CLAIM_PAYMENT_INFORMATION not in claim_payment_line:
        claim_payment_line[TransactionLoops.CLAIM_PAYMENT_INFORMATION] = []

    claim_payment_line[TransactionLoops.CLAIM_PAYMENT_INFORMATION].append(
        {
            "cas_segment": [],
            "nm1_segment": [],
            "ref_segment": [],
            "dtm_segment": [],
            "per_segment": [],
            "amt_segment": [],
            "qty_segment": [],
        }
    )

    loop_data = claim_payment_line[TransactionLoops.CLAIM_PAYMENT_INFORMATION][-1]
    context.set_loop_context(TransactionLoops.CLAIM_PAYMENT_INFORMATION, loop_data)


@match("SVC")
def set_service_payment_information_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the service payment information loop
    """
    claim_payment_information = _get_claim_payment_information(context)

    if TransactionLoops.SERVICE_PAYMENT_INFORMATION not in claim_payment_information:
        claim_payment_information[TransactionLoops.SERVICE_PAYMENT_INFORMATION] = []

    claim_payment_information[TransactionLoops.SERVICE_PAYMENT_INFORMATION].append(
        {
            "dtm_segment": [],
            "cas_segment": [],
            "ref_segment": [],
            "amt_segment": [],
            "qty_segment": [],
            "lq_segment": [],
        }
    )

    loop_data = claim_payment_information[TransactionLoops.SERVICE_PAYMENT_INFORMATION][
        -1
    ]
    context.set_loop_context(TransactionLoops.SERVICE_PAYMENT_INFORMATION, loop_data)


@match("PLB")
def set_plb_footer_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop if the PLB optional segment is encountered.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )


@match("SE")
def set_footer_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )



"""
transaction_set.py

Defines the Health Care Claims Payment 835 005010X221A1 transaction set model.
"""
from x12.models import X12SegmentGroup
from .loops import Header, Footer, Loop1000A, Loop1000B, Loop2000
from pydantic import root_validator
from x12.validators import validate_segment_count
from typing import List, Set


class HealthCareClaimPayment(X12SegmentGroup):
    """
    The Health Care Claim Payment/835 transaction
    """

    header: Header
    loop_1000a: Loop1000A
    loop_1000b: Loop1000B
    loop_2000: List[Loop2000]
    footer: Footer

    _validate_segment_count = root_validator(allow_reuse=True)(validate_segment_count)

    @root_validator()
    def validate_lx_header(cls, values):
        """
        Validates that LX numbers within a transaction set are unique.
        """
        numbers: Set = set()
        for loop in values.get("loop_2000", []):
            n: int = loop.lx_segment.assigned_number
            if n in numbers:
                raise ValueError(f"duplicate assigned_numbers {n}")
            numbers.add(n)

        return values
